import asyncio
import logging
from ssl import SSLContext
from typing import Any, Dict, Literal, Mapping, Optional, Union

import aiohttp
from aiohttp import BasicAuth, ClientWSTimeout, Fingerprint, WSMsgType
from aiohttp.typedefs import LooseHeaders, StrOrURL
from multidict import CIMultiDictProxy

from ...exceptions import TransportConnectionFailed, TransportProtocolError
from ..aiohttp_closed_event import create_aiohttp_closed_event
from .connection import AdapterConnection

log = logging.getLogger("gql.transport.common.adapters.aiohttp")


class AIOHTTPWebSocketsAdapter(AdapterConnection):
    """AdapterConnection implementation using the aiohttp library."""

    def __init__(
        self,
        url: StrOrURL,
        *,
        headers: Optional[LooseHeaders] = None,
        ssl: Optional[Union[SSLContext, Literal[False], Fingerprint]] = None,
        session: Optional[aiohttp.ClientSession] = None,
        client_session_args: Optional[Dict[str, Any]] = None,
        connect_args: Optional[Dict[str, Any]] = None,
        heartbeat: Optional[float] = None,
        auth: Optional[BasicAuth] = None,
        origin: Optional[str] = None,
        params: Optional[Mapping[str, str]] = None,
        proxy: Optional[StrOrURL] = None,
        proxy_auth: Optional[BasicAuth] = None,
        proxy_headers: Optional[LooseHeaders] = None,
        websocket_close_timeout: float = 10.0,
        receive_timeout: Optional[float] = None,
        ssl_close_timeout: Optional[Union[int, float]] = 10,
    ) -> None:
        """Initialize the transport with the given parameters.

        :param url: The GraphQL server URL. Example: 'wss://server.com:PORT/graphql'.
        :param headers: Dict of HTTP Headers.
        :param ssl: SSL validation mode. ``True`` for default SSL check
                      (:func:`ssl.create_default_context` is used),
                      ``False`` for skip SSL certificate validation,
                      :class:`aiohttp.Fingerprint` for fingerprint
                      validation, :class:`ssl.SSLContext` for custom SSL
                      certificate validation.
        :param session: Optional aiohttp opened session.
        :param client_session_args: Dict of extra args passed to
                :class:`aiohttp.ClientSession`
        :param connect_args: Dict of extra args passed to
                :meth:`aiohttp.ClientSession.ws_connect`

        :param float heartbeat: Send low level `ping` message every `heartbeat`
                                seconds and wait `pong` response, close
                                connection if `pong` response is not
                                received. The timer is reset on any data reception.
        :param auth: An object that represents HTTP Basic Authorization.
                     :class:`~aiohttp.BasicAuth` (optional)
        :param str origin: Origin header to send to server(optional)
        :param params: Mapping, iterable of tuple of *key*/*value* pairs or
                       string to be sent as parameters in the query
                       string of the new request. Ignored for subsequent
                       redirected requests (optional)

                       Allowed values are:

                       - :class:`collections.abc.Mapping` e.g. :class:`dict`,
                         :class:`multidict.MultiDict` or
                         :class:`multidict.MultiDictProxy`
                       - :class:`collections.abc.Iterable` e.g. :class:`tuple` or
                         :class:`list`
                       - :class:`str` with preferably url-encoded content
                         (**Warning:** content will not be encoded by *aiohttp*)
        :param proxy: Proxy URL, :class:`str` or :class:`~yarl.URL` (optional)
        :param aiohttp.BasicAuth proxy_auth: an object that represents proxy HTTP
                                             Basic Authorization (optional)
        :param float websocket_close_timeout: Timeout for websocket to close.
                                              ``10`` seconds by default
        :param float receive_timeout: Timeout for websocket to receive
                                      complete message.  ``None`` (unlimited)
                                      seconds by default
        :param ssl_close_timeout: Timeout in seconds to wait for the ssl connection
                                  to close properly
        """
        super().__init__(
            url=str(url),
            connect_args=connect_args,
        )

        self._headers: Optional[LooseHeaders] = headers
        self.ssl: Optional[Union[SSLContext, Literal[False], Fingerprint]] = ssl

        self.session: Optional[aiohttp.ClientSession] = session
        self._using_external_session = True if self.session else False

        if client_session_args is None:
            client_session_args = {}
        self.client_session_args = client_session_args

        self.heartbeat: Optional[float] = heartbeat
        self.auth: Optional[BasicAuth] = auth
        self.origin: Optional[str] = origin
        self.params: Optional[Mapping[str, str]] = params

        self.proxy: Optional[StrOrURL] = proxy
        self.proxy_auth: Optional[BasicAuth] = proxy_auth
        self.proxy_headers: Optional[LooseHeaders] = proxy_headers

        self.websocket_close_timeout: float = websocket_close_timeout
        self.receive_timeout: Optional[float] = receive_timeout

        self.ssl_close_timeout: Optional[Union[int, float]] = ssl_close_timeout

        self.websocket: Optional[aiohttp.ClientWebSocketResponse] = None
        self._response_headers: Optional[CIMultiDictProxy[str]] = None

    async def connect(self) -> None:
        """Connect to the WebSocket server."""

        assert self.websocket is None

        # Create a session if necessary
        if self.session is None:
            client_session_args: Dict[str, Any] = {}

            # Adding custom parameters passed from init
            client_session_args.update(self.client_session_args)  # type: ignore

            self.session = aiohttp.ClientSession(**client_session_args)

        ws_timeout = ClientWSTimeout(
            ws_receive=self.receive_timeout,
            ws_close=self.websocket_close_timeout,
        )

        connect_args: Dict[str, Any] = {
            "url": self.url,
            "headers": self.headers,
            "auth": self.auth,
            "heartbeat": self.heartbeat,
            "origin": self.origin,
            "params": self.params,
            "proxy": self.proxy,
            "proxy_auth": self.proxy_auth,
            "proxy_headers": self.proxy_headers,
            "timeout": ws_timeout,
        }

        if self.subprotocols:
            connect_args["protocols"] = self.subprotocols

        if self.ssl is not None:
            connect_args["ssl"] = self.ssl

        # Adding custom parameters passed from init
        connect_args.update(self.connect_args)

        try:
            self.websocket = await self.session.ws_connect(
                **connect_args,
            )
        except Exception as e:
            raise TransportConnectionFailed("Connect failed") from e

        self._response_headers = self.websocket._response.headers

    async def send(self, message: str) -> None:
        """Send message to the WebSocket server.

        Args:
            message: String message to send

        Raises:
            TransportConnectionFailed: If connection closed
        """
        if self.websocket is None:
            raise TransportConnectionFailed("WebSocket connection is already closed")

        try:
            await self.websocket.send_str(message)
        except Exception as e:
            raise TransportConnectionFailed(
                f"Error trying to send data: {type(e).__name__}"
            ) from e

    async def receive(self) -> str:
        """Receive message from the WebSocket server.

        Returns:
            String message received

        Raises:
            TransportConnectionFailed: If connection closed
            TransportProtocolError: If protocol error or binary data received
        """
        # It is possible that the websocket has been already closed in another task
        if self.websocket is None:
            raise TransportConnectionFailed("Connection is already closed")

        while True:
            # Should not raise any exception:
            # https://docs.aiohttp.org/en/stable/_modules/aiohttp/client_ws.html
            #                                           #ClientWebSocketResponse.receive
            ws_message = await self.websocket.receive()

            # Ignore low-level ping and pong received
            if ws_message.type not in (WSMsgType.PING, WSMsgType.PONG):
                break

        if ws_message.type in (
            WSMsgType.CLOSE,
            WSMsgType.CLOSED,
            WSMsgType.CLOSING,
            WSMsgType.ERROR,
        ):
            raise TransportConnectionFailed("Connection was closed")
        elif ws_message.type is WSMsgType.BINARY:
            raise TransportProtocolError("Binary data received in the websocket")

        assert ws_message.type is WSMsgType.TEXT

        answer: str = ws_message.data

        return answer

    async def _close_session(self) -> None:
        """Close the aiohttp session."""

        assert self.session is not None

        closed_event = create_aiohttp_closed_event(self.session)
        await self.session.close()
        try:
            await asyncio.wait_for(closed_event.wait(), self.ssl_close_timeout)
        except asyncio.TimeoutError:
            pass
        finally:
            self.session = None

    async def close(self) -> None:
        """Close the WebSocket connection."""

        if self.websocket:
            websocket = self.websocket
            self.websocket = None
            try:
                await websocket.close()
            except Exception as exc:  # pragma: no cover
                log.warning("websocket.close() exception: " + repr(exc))

        if self.session and not self._using_external_session:
            await self._close_session()

    @property
    def headers(self) -> Optional[LooseHeaders]:
        """Get the response headers from the WebSocket connection.

        Returns:
            Dictionary of response headers
        """
        if self._headers:
            return self._headers
        return {}

    @property
    def response_headers(self) -> Dict[str, str]:
        """Get the response headers from the WebSocket connection.

        Returns:
            Dictionary of response headers
        """
        if self._response_headers:
            return dict(self._response_headers)
        return {}



"""Utilities to manipulate several python objects."""

from typing import List


# From this response in Stackoverflow
# http://stackoverflow.com/a/19053800/1072990
def to_camel_case(snake_str):
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() if x else "_" for x in components[1:])


def str_first_element(errors: List) -> str:
    try:
        first_error = errors[0]
    except (KeyError, TypeError):
        first_error = errors

    return str(first_error)



import abc
from typing import Any, List

from graphql import ExecutionResult

from ..graphql_request import GraphQLRequest


class Transport(abc.ABC):
    @abc.abstractmethod
    def execute(
        self,
        request: GraphQLRequest,
        *args: Any,
        **kwargs: Any,
    ) -> ExecutionResult:
        """Execute GraphQL query.

        Execute the provided request for either a remote or local GraphQL Schema.

        :param request: GraphQL request as a GraphQLRequest object.
        :return: ExecutionResult
        """
        raise NotImplementedError(
            "Any Transport subclass must implement execute method"
        )  # pragma: no cover

    def execute_batch(
        self,
        reqs: List[GraphQLRequest],
        *args: Any,
        **kwargs: Any,
    ) -> List[ExecutionResult]:
        """Execute multiple GraphQL requests in a batch.

        Execute the provided requests for either a remote or local GraphQL Schema.

        :param reqs: GraphQL requests as a list of GraphQLRequest objects.
        :return: a list of ExecutionResult objects
        """
        raise NotImplementedError(
            "This Transport has not implemented the execute_batch method"
        )

    def connect(self):
        """Establish a session with the transport."""
        pass  # pragma: no cover

    def close(self):
        """Close the transport

        This method doesn't have to be implemented unless the transport would benefit
        from it. This is currently used by the RequestsHTTPTransport transport to close
        the session's connection pool.
        """
        pass  # pragma: no cover



import asyncio
import json
import sys
import warnings
from typing import List

import pytest
from parse import search

from gql import Client, gql
from gql.client import AsyncClientSession
from gql.transport.exceptions import TransportConnectionFailed, TransportServerError

from .conftest import MS, WebSocketServerHelper

# Marking all tests in this file with the websockets marker
pytestmark = pytest.mark.websockets

countdown_server_answer = (
    '{{"type":"next","id":"{query_id}","payload":{{"data":{{"number":{number}}}}}}}'
)

COUNTING_DELAY = 20 * MS
PING_SENDING_DELAY = 50 * MS
PONG_TIMEOUT = 100 * MS

# List which can used to store received messages by the server
logged_messages: List[str] = []


def server_countdown_factory(
    keepalive=False, answer_pings=True, simulate_disconnect=False
):
    async def server_countdown_template(ws):
        import websockets

        logged_messages.clear()

        try:
            await WebSocketServerHelper.send_connection_ack(
                ws, payload="dummy_connection_ack_payload"
            )

            result = await ws.recv()
            logged_messages.append(result)

            json_result = json.loads(result)
            assert json_result["type"] == "subscribe"
            payload = json_result["payload"]
            query = payload["query"]
            query_id = json_result["id"]

            count_found = search("count: {:d}", query)
            count = count_found[0]
            print(f"            Server: Countdown started from: {count}")

            if simulate_disconnect and count == 8:
                await ws.close()

            pong_received: asyncio.Event = asyncio.Event()

            async def counting_coro():
                print("            Server: counting task started")
                try:
                    for number in range(count, -1, -1):
                        await ws.send(
                            countdown_server_answer.format(
                                query_id=query_id, number=number
                            )
                        )
                        await asyncio.sleep(COUNTING_DELAY)
                finally:
                    print("            Server: counting task ended")

            print("            Server: starting counting task")
            counting_task = asyncio.ensure_future(counting_coro())

            async def keepalive_coro():
                print("            Server: keepalive task started")
                try:
                    while True:
                        await asyncio.sleep(PING_SENDING_DELAY)
                        try:
                            # Send a ping
                            await WebSocketServerHelper.send_ping(
                                ws, payload="dummy_ping_payload"
                            )

                            # Wait for a pong
                            try:
                                await asyncio.wait_for(
                                    pong_received.wait(), PONG_TIMEOUT
                                )
                            except asyncio.TimeoutError:
                                print(
                                    "\n            Server: No pong received in time!\n"
                                )
                                break

                            pong_received.clear()

                        except websockets.exceptions.ConnectionClosed:
                            break
                finally:
                    print("            Server: keepalive task ended")

            if keepalive:
                print("            Server: starting keepalive task")
                keepalive_task = asyncio.ensure_future(keepalive_coro())

            async def receiving_coro():
                print("            Server: receiving task started")
                try:
                    nonlocal counting_task
                    while True:

                        try:
                            result = await ws.recv()
                            logged_messages.append(result)
                        except websockets.exceptions.ConnectionClosed:
                            break

                        json_result = json.loads(result)

                        answer_type = json_result["type"]

                        if answer_type == "complete" and json_result["id"] == str(
                            query_id
                        ):
                            print("Cancelling counting task now")
                            counting_task.cancel()
                            if keepalive:
                                print("Cancelling keep alive task now")
                                keepalive_task.cancel()

                        elif answer_type == "ping":
                            if answer_pings:
                                payload = json_result.get("payload", None)
                                await WebSocketServerHelper.send_pong(
                                    ws, payload=payload
                                )

                        elif answer_type == "pong":
                            pong_received.set()
                finally:
                    print("            Server: receiving task ended")
                    if keepalive:
                        keepalive_task.cancel()

            print("            Server: starting receiving task")
            receiving_task = asyncio.ensure_future(receiving_coro())

            try:
                print("            Server: waiting for counting task to complete")
                await counting_task
            except asyncio.CancelledError:
                print("            Server: Now counting task is cancelled")

            print("            Server: sending complete message")
            await WebSocketServerHelper.send_complete(ws, query_id)

            if keepalive:
                print("            Server: cancelling keepalive task")
                keepalive_task.cancel()
                try:
                    await keepalive_task
                except asyncio.CancelledError:
                    print("            Server: Now keepalive task is cancelled")

            print("            Server: waiting for client to close the connection")
            try:
                await asyncio.wait_for(receiving_task, 1000 * MS)
            except asyncio.TimeoutError:
                pass

            print("            Server: cancelling receiving task")
            receiving_task.cancel()

            try:
                await receiving_task
            except asyncio.CancelledError:
                print("            Server: Now receiving task is cancelled")

        except websockets.exceptions.ConnectionClosedOK:
            pass
        except AssertionError as e:
            print(f"\n            Server: Assertion failed: {e!s}\n")
        finally:
            print("            Server: waiting for websocket connection to close")
            await ws.wait_closed()
            print("            Server: connection closed")

    return server_countdown_template


async def server_countdown(ws):

    server = server_countdown_factory()
    await server(ws)


async def server_countdown_keepalive(ws):

    server = server_countdown_factory(keepalive=True)
    await server(ws)


async def server_countdown_dont_answer_pings(ws):

    server = server_countdown_factory(answer_pings=False)
    await server(ws)


async def server_countdown_disconnect(ws):

    server = server_countdown_factory(simulate_disconnect=True)
    await server(ws)


countdown_subscription_str = """
    subscription {{
      countdown (count: {count}) {{
        number
      }}
    }}
"""


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription(client_and_graphqlws_server, subscription_str):

    session, server = client_and_graphqlws_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async for result in session.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_break(
    client_and_graphqlws_server, subscription_str
):

    session, server = client_and_graphqlws_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    generator = session.subscribe(subscription)
    async for result in generator:

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count

        if count <= 5:
            break

        count -= 1

    assert count == 5

    # Using aclose here to make it stop cleanly on pypy
    await generator.aclose()


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_task_cancel(
    client_and_graphqlws_server, subscription_str
):

    session, server = client_and_graphqlws_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    task_cancelled = False

    async def task_coro():
        nonlocal count
        nonlocal task_cancelled

        try:
            async for result in session.subscribe(subscription):

                number = result["number"]
                print(f"Number received: {number}")

                assert number == count

                count -= 1
        except asyncio.CancelledError:
            print("Inside task cancelled")
            task_cancelled = True

    task = asyncio.ensure_future(task_coro())

    async def cancel_task_coro():
        nonlocal task

        await asyncio.sleep(5.5 * COUNTING_DELAY)

        task.cancel()

    cancel_task = asyncio.ensure_future(cancel_task_coro())

    await asyncio.gather(task, cancel_task)

    assert count > 0
    assert task_cancelled is True


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_close_transport(
    client_and_graphqlws_server, subscription_str
):

    session, server = client_and_graphqlws_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async def task_coro():
        nonlocal count
        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count

            count -= 1

    task = asyncio.ensure_future(task_coro())

    async def close_transport_task_coro():
        nonlocal task

        await asyncio.sleep(5.5 * COUNTING_DELAY)

        await session.transport.close()

    close_transport_task = asyncio.ensure_future(close_transport_task_coro())

    await asyncio.gather(task, close_transport_task)

    assert count > 0


async def server_countdown_close_connection_in_middle(ws):
    await WebSocketServerHelper.send_connection_ack(ws)

    result = await ws.recv()
    json_result = json.loads(result)
    assert json_result["type"] == "subscribe"
    payload = json_result["payload"]
    query = payload["query"]
    query_id = json_result["id"]

    count_found = search("count: {:d}", query)
    count = count_found[0]
    stopping_before = count // 2
    print(f"Countdown started from: {count}, stopping server before {stopping_before}")
    for number in range(count, stopping_before, -1):
        await ws.send(countdown_server_answer.format(query_id=query_id, number=number))
        await asyncio.sleep(COUNTING_DELAY)

    print("Closing server while subscription is still running now")
    await ws.close()
    await ws.wait_closed()
    print("Server is now closed")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_close_connection_in_middle], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_server_connection_closed(
    client_and_graphqlws_server, subscription_str
):
    session, server = client_and_graphqlws_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    with pytest.raises(TransportConnectionFailed):

        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count

            count -= 1


@pytest.mark.asyncio
@pytest.mark.parametrize("graphqlws_server", [server_countdown], indirect=True)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_with_operation_name(
    client_and_graphqlws_server, subscription_str
):

    session, server = client_and_graphqlws_server

    count = 10
    subscription = gql(subscription_str.format(count=count))
    subscription.operation_name = "CountdownSubscription"

    async for result in session.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1

    # Check that the query contains the operationName
    assert '"operationName": "CountdownSubscription"' in logged_messages[0]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_with_keepalive(
    client_and_graphqlws_server, subscription_str
):

    session, server = client_and_graphqlws_server

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async for result in session.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1
    assert "ping" in session.transport.payloads
    assert session.transport.payloads["ping"] == "dummy_ping_payload"
    assert (
        session.transport.payloads["connection_ack"] == "dummy_connection_ack_payload"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_with_keepalive_with_timeout_ok(
    graphqlws_server, subscription_str
):

    from gql.transport.websockets import WebsocketsTransport

    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(url=url, keep_alive_timeout=(5 * COUNTING_DELAY))

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count
            count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_with_keepalive_with_timeout_nok(
    graphqlws_server, subscription_str
):

    from gql.transport.websockets import WebsocketsTransport

    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(url=url, keep_alive_timeout=(COUNTING_DELAY / 2))

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        with pytest.raises(TransportServerError) as exc_info:
            async for result in session.subscribe(subscription):

                number = result["number"]
                print(f"Number received: {number}")

                assert number == count
                count -= 1

        assert "No keep-alive message has been received" in str(exc_info.value)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_with_ping_interval_ok(
    graphqlws_server, subscription_str
):

    from gql.transport.websockets import WebsocketsTransport

    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(
        url=url,
        ping_interval=(5 * COUNTING_DELAY),
        pong_timeout=(4 * COUNTING_DELAY),
    )

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count
            count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_dont_answer_pings], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_with_ping_interval_nok(
    graphqlws_server, subscription_str
):

    from gql.transport.websockets import WebsocketsTransport

    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(url=url, ping_interval=(5 * COUNTING_DELAY))

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        with pytest.raises(TransportServerError) as exc_info:
            async for result in session.subscribe(subscription):

                number = result["number"]
                print(f"Number received: {number}")

                assert number == count
                count -= 1

        assert "No pong received" in str(exc_info.value)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_manual_pings_with_payload(
    graphqlws_server, subscription_str
):

    from gql.transport.websockets import WebsocketsTransport

    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:
        async for result in session.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            payload = {"count_received": count}

            await transport.send_ping(payload=payload)

            await asyncio.wait_for(transport.pong_received.wait(), 10000 * MS)

            transport.pong_received.clear()

            assert transport.payloads["pong"] == payload

            assert number == count
            count -= 1

    assert count == -1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_manual_pong_answers_with_payload(
    graphqlws_server, subscription_str
):

    from gql.transport.websockets import WebsocketsTransport

    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(url=url, answer_pings=False)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    async with client as session:

        async def answer_ping_coro():
            while True:
                await transport.ping_received.wait()
                transport.ping_received.clear()
                await transport.send_pong(payload={"some": "data"})

        answer_ping_task = asyncio.ensure_future(answer_ping_coro())

        try:
            async for result in session.subscribe(subscription):

                number = result["number"]
                print(f"Number received: {number}")

                assert number == count
                count -= 1

        finally:
            answer_ping_task.cancel()

    assert count == -1


@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
def test_graphqlws_subscription_sync(graphqlws_server, subscription_str):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    for result in client.subscribe(subscription):

        number = result["number"]
        print(f"Number received: {number}")

        assert number == count
        count -= 1

    assert count == -1


@pytest.mark.skipif(sys.platform.startswith("win"), reason="test failing on windows")
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
def test_graphqlws_subscription_sync_graceful_shutdown(
    graphqlws_server, subscription_str
):
    """Note: this test will simulate a control-C happening while a sync subscription
    is in progress. To do that we will throw a KeyboardInterrupt exception inside
    the subscription async generator.

    The code should then do a clean close:
      - send stop messages for each active query
      - send a connection_terminate message
    Then the KeyboardInterrupt will be reraise (to warn potential user code)

    This test does not work on Windows but the behaviour with Windows is correct.
    """
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 10
    subscription = gql(subscription_str.format(count=count))

    with pytest.raises(KeyboardInterrupt):
        for result in client.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count

            if count == 5:

                # Simulate a KeyboardInterrupt in the generator
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", message="There is no current event loop"
                    )
                    assert isinstance(client.session, AsyncClientSession)
                    asyncio.ensure_future(
                        client.session._generator.athrow(KeyboardInterrupt)
                    )

            count -= 1

    assert count == 4

    # Check that the server received a connection_terminate message last
    # assert logged_messages.pop() == '{"type": "connection_terminate"}'


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_keepalive], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
async def test_graphqlws_subscription_running_in_thread(
    graphqlws_server, subscription_str, run_sync_test
):
    from gql.transport.websockets import WebsocketsTransport

    def test_code():
        path = "/graphql"
        url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
        transport = WebsocketsTransport(url=url)

        client = Client(transport=transport)

        count = 10
        subscription = gql(subscription_str.format(count=count))

        for result in client.subscribe(subscription):

            number = result["number"]
            print(f"Number received: {number}")

            assert number == count
            count -= 1

        assert count == -1

    await run_sync_test(graphqlws_server, test_code)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "graphqlws_server", [server_countdown_disconnect], indirect=True
)
@pytest.mark.parametrize("subscription_str", [countdown_subscription_str])
@pytest.mark.parametrize("execute_instead_of_subscribe", [False, True])
async def test_graphqlws_subscription_reconnecting_session(
    graphqlws_server, subscription_str, execute_instead_of_subscribe
):

    from gql.transport.websockets import WebsocketsTransport

    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(url=url)

    client = Client(transport=transport)

    count = 8
    subscription_with_disconnect = gql(subscription_str.format(count=count))

    count = 10
    subscription = gql(subscription_str.format(count=count))

    session = await client.connect_async(
        reconnecting=True, retry_connect=False, retry_execute=False
    )

    # First we make a query or subscription which will cause a disconnect
    # in the backend (count=8)
    try:
        if execute_instead_of_subscribe:
            print("\nEXECUTION_1\n")
            await session.execute(subscription_with_disconnect)
        else:
            print("\nSUBSCRIPTION_1_WITH_DISCONNECT\n")
            async for result in session.subscribe(subscription_with_disconnect):
                pass
    except TransportConnectionFailed:
        pass

    # Wait for disconnect
    for i in range(200):
        await asyncio.sleep(1 * MS)
        if not transport._connected:
            print(f"\nDisconnected in {i+1} MS")
            break

    # Wait for reconnect
    for i in range(200):
        await asyncio.sleep(1 * MS)
        if transport._connected:
            print(f"\nConnected again in {i+1} MS")
            break

    assert transport._connected is True

    # Then after the reconnection, we make a query or a subscription
    if execute_instead_of_subscribe:
        print("\nEXECUTION_2\n")
        result = await session.execute(subscription)
        assert result["number"] == 10
    else:
        print("\nSUBSCRIPTION_2\n")
        generator = session.subscribe(subscription)
        async for result in generator:
            number = result["number"]
            print(f"Number received: {number}")

            assert number == count
            count -= 1

        await generator.aclose()

        assert count == -1

    # Close the reconnecting session
    await client.close_async()

    # Wait for disconnect
    for i in range(200):
        await asyncio.sleep(1 * MS)
        if not transport._connected:
            print(f"\nDisconnected in {i+1} MS")
            break

    assert transport._connected is False



import pytest

from gql import Client
from gql.dsl import DSLQuery, DSLSchema, dsl_gql
from tests.nested_input.schema import NestedInputSchema


@pytest.fixture
def ds():
    return DSLSchema(NestedInputSchema)


@pytest.fixture
def client():
    return Client(schema=NestedInputSchema)


def test_nested_input(ds, client):
    query = dsl_gql(DSLQuery(ds.Query.echo.args(nested={"foo": 1})))
    assert client.execute(query) == {"echo": '{"foo": 1}'}


def test_nested_input_2(ds, client):
    query = dsl_gql(
        DSLQuery(ds.Query.echo.args(nested={"foo": 1, "child": {"foo": 2}}))
    )
    assert client.execute(query) == {"echo": '{"foo": 1, "child": {"foo": 2}}'}


def test_nested_input_3(ds, client):
    query = dsl_gql(
        DSLQuery(
            ds.Query.echo.args(
                nested={"foo": 1, "child": {"foo": 2, "child": {"foo": 3}}}
            )
        )
    )
    assert client.execute(query) == {
        "echo": '{"foo": 1, "child": {"foo": 2, "child": {"foo": 3}}}'
    }



from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

transport = RequestsHTTPTransport(
    url="https://countries.trevorblades.com/",
    verify=True,
    retries=3,
)

client = Client(transport=transport, fetch_schema_from_transport=True)

query = gql(
    """
    query getContinents {
      continents {
        code
        name
      }
    }
"""
)

result = client.execute(query)
print(result)



import asyncio
import json
import sys
from typing import Any, Dict, Mapping

import pytest

from gql import Client, gql
from gql.transport.exceptions import (
    TransportAlreadyConnected,
    TransportConnectionFailed,
    TransportQueryError,
    TransportServerError,
)

from .conftest import MS, WebSocketServerHelper, get_localhost_ssl_context_client

# Marking all tests in this file with the aiohttp AND websockets marker
pytestmark = pytest.mark.aiohttp

query1_str = """
    query getContinents {
      continents {
        code
        name
      }
    }
"""

query1_server_answer_data = (
    '{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}'
)

query1_server_answer = (
    '{{"type":"data","id":"{query_id}","payload":{{"data":{{"continents":['
    '{{"code":"AF","name":"Africa"}},{{"code":"AN","name":"Antarctica"}},'
    '{{"code":"AS","name":"Asia"}},{{"code":"EU","name":"Europe"}},'
    '{{"code":"NA","name":"North America"}},{{"code":"OC","name":"Oceania"}},'
    '{{"code":"SA","name":"South America"}}]}}}}}}'
)

server1_answers = [
    query1_server_answer,
]


@pytest.mark.asyncio
@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
async def test_aiohttp_websocket_starting_client_in_context_manager(aiohttp_ws_server):

    server = aiohttp_ws_server
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(
        url=url,
        websocket_close_timeout=10,
        headers={"test": "1234"},
    )

    assert transport.response_headers == {}
    assert isinstance(transport.headers, Mapping)
    assert transport.headers["test"] == "1234"  # type: ignore

    async with Client(transport=transport) as session:

        query1 = gql(query1_str)

        result = await session.execute(query1)

        print("Client received:", result)

        # Verify result
        assert isinstance(result, Dict)

        continents = result["continents"]
        africa = continents[0]

        assert africa["code"] == "AF"

        # Checking response headers are saved in the transport
        assert hasattr(transport, "response_headers")
        assert isinstance(transport.response_headers, Mapping)
        assert transport.response_headers["dummy"] == "test1234"

    # Check client is disconnect here
    assert transport._connected is False


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize("ws_ssl_server", [server1_answers], indirect=True)
@pytest.mark.parametrize("ssl_close_timeout", [0, 10])
@pytest.mark.parametrize("verify_https", ["disabled", "cert_provided"])
async def test_aiohttp_websocket_using_ssl_connection(
    ws_ssl_server, ssl_close_timeout, verify_https
):

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    server = ws_ssl_server

    url = f"wss://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    extra_args = {}

    if verify_https == "cert_provided":
        _, ssl_context = get_localhost_ssl_context_client()

        extra_args["ssl"] = ssl_context
    elif verify_https == "disabled":
        extra_args["ssl"] = False

    transport = AIOHTTPWebsocketsTransport(
        url=url,
        ssl_close_timeout=ssl_close_timeout,
        **extra_args,
    )

    async with Client(transport=transport) as session:

        query1 = gql(query1_str)

        result = await session.execute(query1)

        print("Client received:", result)

        # Verify result
        assert isinstance(result, Dict)

        continents = result["continents"]
        africa = continents[0]

        assert africa["code"] == "AF"

    # Check client is disconnect here
    assert transport._connected is False


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize("ws_ssl_server", [server1_answers], indirect=True)
@pytest.mark.parametrize("ssl_close_timeout", [10])
@pytest.mark.parametrize("verify_https", ["explicitely_enabled", "default"])
async def test_aiohttp_websocket_using_ssl_connection_self_cert_fail(
    ws_ssl_server, ssl_close_timeout, verify_https
):

    from aiohttp.client_exceptions import ClientConnectorCertificateError

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    server = ws_ssl_server

    url = f"wss://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    extra_args: Dict[str, Any] = {}

    if verify_https == "explicitely_enabled":
        extra_args["ssl"] = True

    transport = AIOHTTPWebsocketsTransport(
        url=url,
        ssl_close_timeout=ssl_close_timeout,
        **extra_args,
    )

    if verify_https == "explicitely_enabled":
        assert transport.ssl is True

    with pytest.raises(TransportConnectionFailed) as exc_info:
        async with Client(transport=transport) as session:

            query1 = gql(query1_str)

            await session.execute(query1)

    cause = exc_info.value.__cause__

    assert isinstance(cause, ClientConnectorCertificateError)

    expected_error = "certificate verify failed: self-signed certificate"

    assert expected_error in str(cause)

    # Check client is disconnect here
    assert transport._connected is False


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize("server", [server1_answers], indirect=True)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_simple_query(aiohttp_client_and_server, query_str):

    session, server = aiohttp_client_and_server

    query = gql(query_str)

    result = await session.execute(query)

    print("Client received:", result)


server1_two_answers_in_series = [
    query1_server_answer,
    query1_server_answer,
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "aiohttp_ws_server", [server1_two_answers_in_series], indirect=True
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_two_queries_in_series(
    aiohttp_client_and_aiohttp_ws_server, query_str
):

    session, server = aiohttp_client_and_aiohttp_ws_server

    query = gql(query_str)

    result1 = await session.execute(query)

    print("Query1 received:", result1)

    result2 = await session.execute(query)

    print("Query2 received:", result2)

    assert result1 == result2


async def server1_two_queries_in_parallel(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    result = await ws.recv()
    print(f"Server received: {result}", file=sys.stderr)
    result = await ws.recv()
    print(f"Server received: {result}", file=sys.stderr)
    await ws.send(query1_server_answer.format(query_id=1))
    await ws.send(query1_server_answer.format(query_id=2))
    await WebSocketServerHelper.send_complete(ws, 1)
    await WebSocketServerHelper.send_complete(ws, 2)
    await WebSocketServerHelper.wait_connection_terminate(ws)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize("server", [server1_two_queries_in_parallel], indirect=True)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_two_queries_in_parallel(
    aiohttp_client_and_server, query_str
):

    session, server = aiohttp_client_and_server

    query = gql(query_str)

    result1 = None
    result2 = None

    async def task1_coro():
        nonlocal result1
        result1 = await session.execute(query)

    async def task2_coro():
        nonlocal result2
        result2 = await session.execute(query)

    task1 = asyncio.ensure_future(task1_coro())
    task2 = asyncio.ensure_future(task2_coro())

    await asyncio.gather(task1, task2)

    print("Query1 received:", result1)
    print("Query2 received:", result2)

    assert result1 == result2


async def server_closing_while_we_are_doing_something_else(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    result = await ws.recv()
    print(f"Server received: {result}", file=sys.stderr)
    await ws.send(query1_server_answer.format(query_id=1))
    await WebSocketServerHelper.send_complete(ws, 1)
    await asyncio.sleep(1 * MS)

    # Closing server after first query
    await ws.close()


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize(
    "server", [server_closing_while_we_are_doing_something_else], indirect=True
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_server_closing_after_first_query(
    aiohttp_client_and_server, query_str
):

    session, server = aiohttp_client_and_server

    query = gql(query_str)

    # First query is working
    await session.execute(query)

    # Then we do other things
    await asyncio.sleep(10 * MS)

    # Now the server is closed but we don't know it yet, we have to send a query
    # to notice it and to receive the exception
    with pytest.raises(TransportConnectionFailed):
        await session.execute(query)


ignore_invalid_id_answers = [
    query1_server_answer,
    '{"type":"complete","id": "55"}',
    query1_server_answer,
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "aiohttp_ws_server", [ignore_invalid_id_answers], indirect=True
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_ignore_invalid_id(
    aiohttp_client_and_aiohttp_ws_server, query_str
):

    session, server = aiohttp_client_and_aiohttp_ws_server

    query = gql(query_str)

    # First query is working
    await session.execute(query)

    # Second query gets no answer -> raises
    with pytest.raises(TransportQueryError):
        await session.execute(query)

    # Third query is working
    await session.execute(query)


async def assert_client_is_working(session):
    query1 = gql(query1_str)

    result = await session.execute(query1)

    print("Client received:", result)

    # Verify result
    assert isinstance(result, Dict)

    continents = result["continents"]
    africa = continents[0]

    assert africa["code"] == "AF"


@pytest.mark.asyncio
@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
async def test_aiohttp_websocket_multiple_connections_in_series(aiohttp_ws_server):

    server = aiohttp_ws_server

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url)

    async with Client(transport=transport) as session:
        await assert_client_is_working(session)

    # Check client is disconnect here
    assert transport._connected is False

    async with Client(transport=transport) as session:
        await assert_client_is_working(session)

    # Check client is disconnect here
    assert transport._connected is False


@pytest.mark.asyncio
@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
async def test_aiohttp_websocket_multiple_connections_in_parallel(aiohttp_ws_server):

    server = aiohttp_ws_server

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    async def task_coro():
        transport = AIOHTTPWebsocketsTransport(url=url)
        async with Client(transport=transport) as session:
            await assert_client_is_working(session)

    task1 = asyncio.ensure_future(task_coro())
    task2 = asyncio.ensure_future(task_coro())

    await asyncio.gather(task1, task2)


@pytest.mark.asyncio
@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
async def test_aiohttp_websocket_trying_to_connect_to_already_connected_transport(
    aiohttp_ws_server,
):
    server = aiohttp_ws_server

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url)
    async with Client(transport=transport) as session:
        await assert_client_is_working(session)

        with pytest.raises(TransportAlreadyConnected):
            async with Client(transport=transport):
                pass


async def server_with_authentication_in_connection_init_payload(ws):
    # Wait the connection_init message
    init_message_str = await ws.recv()
    init_message = json.loads(init_message_str)
    payload = init_message["payload"]

    if "Authorization" in payload:
        if payload["Authorization"] == 12345:
            await ws.send('{"type":"connection_ack"}')

            result = await ws.recv()
            print(f"Server received: {result}", file=sys.stderr)
            await ws.send(query1_server_answer.format(query_id=1))
            await WebSocketServerHelper.send_complete(ws, 1)
        else:
            await ws.send(
                '{"type":"connection_error", "payload": "Invalid Authorization token"}'
            )
    else:
        await ws.send(
            '{"type":"connection_error", "payload": "No Authorization token"}'
        )

    await ws.close()


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize(
    "server", [server_with_authentication_in_connection_init_payload], indirect=True
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_connect_success_with_authentication_in_connection_init(
    server, query_str
):
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    init_payload = {"Authorization": 12345}

    transport = AIOHTTPWebsocketsTransport(url=url, init_payload=init_payload)

    async with Client(transport=transport) as session:

        query1 = gql(query_str)

        result = await session.execute(query1)

        print("Client received:", result)

        # Verify result
        assert isinstance(result, Dict)

        continents = result["continents"]
        africa = continents[0]

        assert africa["code"] == "AF"


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize(
    "server", [server_with_authentication_in_connection_init_payload], indirect=True
)
@pytest.mark.parametrize("query_str", [query1_str])
@pytest.mark.parametrize("init_payload", [{}, {"Authorization": "invalid_code"}])
async def test_aiohttp_websocket_connect_failed_with_authentication_in_connection_init(
    server, query_str, init_payload
):
    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url, init_payload=init_payload)

    for _ in range(2):
        with pytest.raises(TransportServerError):
            async with Client(transport=transport) as session:
                query1 = gql(query_str)

                await session.execute(query1)

        assert transport.adapter.session is None
        assert transport._connected is False


@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
def test_aiohttp_websocket_execute_sync(aiohttp_ws_server):
    server = aiohttp_ws_server

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = AIOHTTPWebsocketsTransport(url=url)

    client = Client(transport=transport)

    query1 = gql(query1_str)

    result = client.execute(query1)

    print("Client received:", result)

    # Verify result
    assert isinstance(result, Dict)

    continents = result["continents"]
    africa = continents[0]

    assert africa["code"] == "AF"

    # Execute sync a second time
    result = client.execute(query1)

    print("Client received:", result)

    # Verify result
    assert isinstance(result, Dict)

    continents = result["continents"]
    africa = continents[0]

    assert africa["code"] == "AF"

    # Check client is disconnect here
    assert transport._connected is False


@pytest.mark.asyncio
@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
async def test_aiohttp_websocket_add_extra_parameters_to_connect(aiohttp_ws_server):

    server = aiohttp_ws_server

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"

    # Increase max payload size
    transport = AIOHTTPWebsocketsTransport(
        url=url,
        connect_args={
            "max_msg_size": 2**21,
        },
    )

    query = gql(query1_str)

    async with Client(transport=transport) as session:
        await session.execute(query)


async def server_sending_keep_alive_before_connection_ack(ws):
    await WebSocketServerHelper.send_keepalive(ws)
    await WebSocketServerHelper.send_keepalive(ws)
    await WebSocketServerHelper.send_keepalive(ws)
    await WebSocketServerHelper.send_keepalive(ws)
    await WebSocketServerHelper.send_connection_ack(ws)
    result = await ws.recv()
    print(f"Server received: {result}", file=sys.stderr)
    await ws.send(query1_server_answer.format(query_id=1))
    await WebSocketServerHelper.send_complete(ws, 1)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.parametrize(
    "server", [server_sending_keep_alive_before_connection_ack], indirect=True
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_non_regression_bug_108(
    aiohttp_client_and_server, query_str
):

    # This test will check that we now ignore keepalive message
    # arriving before the connection_ack
    # See bug #108

    session, server = aiohttp_client_and_server

    query = gql(query_str)

    result = await session.execute(query)

    print("Client received:", result)

    continents = result["continents"]
    africa = continents[0]

    assert africa["code"] == "AF"


@pytest.mark.asyncio
@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
@pytest.mark.parametrize("transport_arg", [[], ["--transport=aiohttp_websockets"]])
async def test_aiohttp_websocket_using_cli(
    aiohttp_ws_server, transport_arg, monkeypatch, capsys
):
    """
    Note: depending on the transport_arg parameter, if there is no transport argument,
    then we will use WebsocketsTransport if the websockets dependency is installed,
    or AIOHTTPWebsocketsTransport if that is not the case.
    """

    server = aiohttp_ws_server

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    import io
    import json

    from gql.cli import get_parser, main

    parser = get_parser(with_examples=True)
    args = parser.parse_args([url, *transport_arg])

    # Monkeypatching sys.stdin to simulate getting the query
    # via the standard input
    monkeypatch.setattr("sys.stdin", io.StringIO(query1_str))

    # Flush captured output
    captured = capsys.readouterr()

    exit_code = await main(args)

    assert exit_code == 0

    # Check that the result has been printed on stdout
    captured = capsys.readouterr()
    captured_out = str(captured.out).strip()

    expected_answer = json.loads(query1_server_answer_data)
    print(f"Captured: {captured_out}")
    received_answer = json.loads(captured_out)

    assert received_answer == expected_answer


query1_server_answer_with_extensions = (
    '{{"type":"data","id":"{query_id}","payload":{{"data":{{"continents":['
    '{{"code":"AF","name":"Africa"}},{{"code":"AN","name":"Antarctica"}},'
    '{{"code":"AS","name":"Asia"}},{{"code":"EU","name":"Europe"}},'
    '{{"code":"NA","name":"North America"}},{{"code":"OC","name":"Oceania"}},'
    '{{"code":"SA","name":"South America"}}]}},'
    '"extensions": {{"key1": "val1"}}}}}}'
)

server1_answers_with_extensions = [
    query1_server_answer_with_extensions,
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "aiohttp_ws_server", [server1_answers_with_extensions], indirect=True
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_aiohttp_websocket_simple_query_with_extensions(
    aiohttp_client_and_aiohttp_ws_server, query_str
):

    session, server = aiohttp_client_and_aiohttp_ws_server

    query = gql(query_str)

    execution_result = await session.execute(query, get_execution_result=True)

    assert execution_result.extensions["key1"] == "val1"


@pytest.mark.asyncio
@pytest.mark.parametrize("aiohttp_ws_server", [server1_answers], indirect=True)
async def test_aiohttp_websocket_connector_owner_false(aiohttp_ws_server):

    server = aiohttp_ws_server

    from aiohttp import TCPConnector

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    connector = TCPConnector()
    transport = AIOHTTPWebsocketsTransport(
        url=url,
        client_session_args={
            "connector": connector,
            "connector_owner": False,
        },
    )

    for _ in range(2):
        async with Client(transport=transport) as session:

            query1 = gql(query1_str)

            result = await session.execute(query1)

            print("Client received:", result)

            assert isinstance(result, Dict)

            continents = result["continents"]
            africa = continents[0]

            assert africa["code"] == "AF"

    # Check client is disconnect here
    assert transport._connected is False

    await connector.close()



import logging
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple, Union, cast

from graphql import (
    IDLE,
    REMOVE,
    DocumentNode,
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLError,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLType,
    InlineFragmentNode,
    NameNode,
    Node,
    OperationDefinitionNode,
    SelectionSetNode,
    TypeInfo,
    TypeInfoVisitor,
    Visitor,
    is_leaf_type,
    print_ast,
    visit,
)
from graphql.language.visitor import VisitorActionEnum
from graphql.pyutils import inspect

log = logging.getLogger(__name__)

# Equivalent to QUERY_DOCUMENT_KEYS but only for fields interesting to
# visit to parse the results
RESULT_DOCUMENT_KEYS: Dict[str, Tuple[str, ...]] = {
    "document": ("definitions",),
    "operation_definition": ("selection_set",),
    "selection_set": ("selections",),
    "field": ("selection_set",),
    "inline_fragment": ("selection_set",),
    "fragment_definition": ("selection_set",),
}


def _ignore_non_null(type_: GraphQLType) -> GraphQLType:
    """Removes the GraphQLNonNull wrappings around types."""
    if isinstance(type_, GraphQLNonNull):
        return type_.of_type
    else:
        return type_


def _get_fragment(document, fragment_name):
    """Returns a fragment from the document."""
    for definition in document.definitions:
        if isinstance(definition, FragmentDefinitionNode):
            if definition.name.value == fragment_name:
                return definition

    raise GraphQLError(f'Fragment "{fragment_name}" not found in document!')


class ParseResultVisitor(Visitor):
    def __init__(
        self,
        schema: GraphQLSchema,
        document: DocumentNode,
        node: Node,
        result: Dict[str, Any],
        type_info: TypeInfo,
        visit_fragment: bool = False,
        inside_list_level: int = 0,
        operation_name: Optional[str] = None,
    ):
        """Recursive Implementation of a Visitor class to parse results
        correspondind to a schema and a document.

        Using a TypeInfo class to get the node types during traversal.

        If we reach a list in the results, then we parse each
        item of the list recursively, traversing the same nodes
        of the query again.

        During traversal, we keep the current position in the result
        in the result_stack field.

        Alongside the field type, we calculate the "result type"
        which is computed from the field type and the current
        recursive level we are for this field
        (:code:`inside_list_level` argument).
        """
        self.schema: GraphQLSchema = schema
        self.document: DocumentNode = document
        self.node: Node = node
        self.result: Dict[str, Any] = result
        self.type_info: TypeInfo = type_info
        self.visit_fragment: bool = visit_fragment
        self.inside_list_level = inside_list_level
        self.operation_name = operation_name

        self.result_stack: List[Any] = []

        super().__init__()

    @property
    def current_result(self):
        try:
            return self.result_stack[-1]
        except IndexError:
            return self.result

    @staticmethod
    def leave_document(node: DocumentNode, *_args: Any) -> Dict[str, Any]:
        results = cast(List[Dict[str, Any]], node.definitions)
        return {k: v for result in results for k, v in result.items()}

    def enter_operation_definition(
        self, node: OperationDefinitionNode, *_args: Any
    ) -> Union[None, VisitorActionEnum]:

        if self.operation_name is not None:
            if not hasattr(node.name, "value"):
                return REMOVE  # pragma: no cover

            node.name = cast(NameNode, node.name)

            if node.name.value != self.operation_name:
                log.debug(f"SKIPPING operation {node.name.value}")
                return REMOVE

        return IDLE

    @staticmethod
    def leave_operation_definition(
        node: OperationDefinitionNode, *_args: Any
    ) -> Dict[str, Any]:
        selections = cast(List[Dict[str, Any]], node.selection_set)
        return {k: v for s in selections for k, v in s.items()}

    @staticmethod
    def leave_selection_set(node: SelectionSetNode, *_args: Any) -> Dict[str, Any]:
        partial_results = cast(Dict[str, Any], node.selections)
        return partial_results

    @staticmethod
    def in_first_field(path):
        return path.count("selections") <= 1

    def get_current_result_type(self, path):
        field_type = self.type_info.get_type()

        list_level = self.inside_list_level

        assert field_type is not None

        result_type = _ignore_non_null(field_type)

        if self.in_first_field(path):

            while list_level > 0:
                assert isinstance(result_type, GraphQLList)
                result_type = _ignore_non_null(result_type.of_type)

                list_level -= 1

        return result_type

    def enter_field(
        self,
        node: FieldNode,
        key: str,
        parent: Node,
        path: List[Node],
        ancestors: List[Node],
    ) -> Union[None, VisitorActionEnum, Dict[str, Any]]:

        name = node.alias.value if node.alias else node.name.value

        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"Enter field {name}")
            log.debug(f"  path={path!r}")
            log.debug(f"  current_result={self.current_result!r}")

        if self.current_result is None:
            # Result was null for this field -> remove
            return REMOVE

        elif isinstance(self.current_result, Mapping):

            try:
                result_value = self.current_result[name]
            except KeyError:
                # Key not found in result.
                # Should never happen in theory with a correct GraphQL backend
                # Silently ignoring this field
                log.debug(f"  Key {name} not found in result --> REMOVE")
                return REMOVE

            log.debug(f"  result_value={result_value}")

            # We get the field_type from type_info
            field_type = self.type_info.get_type()

            # We calculate a virtual "result type" depending on our recursion level.
            result_type = self.get_current_result_type(path)

            # If the result for this field is a list, then we need
            # to recursively visit the same node multiple times for each
            # item in the list.
            if (
                not isinstance(result_value, Mapping)
                and isinstance(result_value, Iterable)
                and not isinstance(result_value, str)
                and not is_leaf_type(result_type)
            ):

                # Finding out the inner type of the list
                inner_type = _ignore_non_null(result_type.of_type)

                if log.isEnabledFor(logging.DEBUG):
                    log.debug("  List detected:")
                    log.debug(f"    field_type={inspect(field_type)}")
                    log.debug(f"    result_type={inspect(result_type)}")
                    log.debug(f"    inner_type={inspect(inner_type)}\n")

                visits: List[Dict[str, Any]] = []

                # Get parent type
                initial_type = self.type_info.get_parent_type()
                assert isinstance(
                    initial_type, (GraphQLObjectType, GraphQLInterfaceType)
                )

                # Get parent SelectionSet node
                selection_set_node = ancestors[-1]
                assert isinstance(selection_set_node, SelectionSetNode)

                # Keep only the current node in a new selection set node
                new_node = SelectionSetNode(selections=[node])

                for item in result_value:

                    new_result = {name: item}

                    if log.isEnabledFor(logging.DEBUG):
                        log.debug(f"      recursive new_result={new_result}")
                        log.debug(f"      recursive ast={print_ast(node)}")
                        log.debug(f"      recursive path={path!r}")
                        log.debug(f"      recursive initial_type={initial_type!r}\n")

                    if self.in_first_field(path):
                        inside_list_level = self.inside_list_level + 1
                    else:
                        inside_list_level = 1

                    inner_visit = parse_result_recursive(
                        self.schema,
                        self.document,
                        new_node,
                        new_result,
                        initial_type=initial_type,
                        inside_list_level=inside_list_level,
                    )
                    log.debug(f"      recursive result={inner_visit}\n")

                    inner_visit = cast(List[Dict[str, Any]], inner_visit)
                    visits.append(inner_visit[0][name])

                result_value = {name: visits}
                log.debug(f"    recursive visits final result = {result_value}\n")
                return result_value

            # If the result for this field is not a list, then add it
            # to the result stack so that it becomes the current_value
            # for the next inner fields
            self.result_stack.append(result_value)

            return IDLE

        raise GraphQLError(
            f"Invalid result for container of field {name}: {self.current_result!r}"
        )

    def leave_field(
        self,
        node: FieldNode,
        key: str,
        parent: Node,
        path: List[Node],
        ancestors: List[Node],
    ) -> Dict[str, Any]:

        name = cast(str, node.alias.value if node.alias else node.name.value)

        log.debug(f"Leave field {name}")

        if self.current_result is None:

            return_value = None

        elif node.selection_set is None:

            field_type = self.type_info.get_type()
            result_type = self.get_current_result_type(path)

            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"  field type of {name} is {inspect(field_type)}")
                log.debug(f"  result type of {name} is {inspect(result_type)}")

            assert is_leaf_type(result_type)

            # Finally parsing a single scalar using the parse_value method
            return_value = result_type.parse_value(self.current_result)
        else:

            partial_results = cast(List[Dict[str, Any]], node.selection_set)

            return_value = {k: v for pr in partial_results for k, v in pr.items()}

        # Go up a level in the result stack
        self.result_stack.pop()

        log.debug(f"Leave field {name}: returning {return_value}")

        return {name: return_value}

    # Fragments

    def enter_fragment_definition(
        self, node: FragmentDefinitionNode, *_args: Any
    ) -> Union[None, VisitorActionEnum]:

        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"Enter fragment definition {node.name.value}.")
            log.debug(f"visit_fragment={self.visit_fragment!s}")

        if self.visit_fragment:
            return IDLE
        else:
            return REMOVE

    @staticmethod
    def leave_fragment_definition(
        node: FragmentDefinitionNode, *_args: Any
    ) -> Dict[str, Any]:

        selections = cast(List[Dict[str, Any]], node.selection_set)
        return {k: v for s in selections for k, v in s.items()}

    def leave_fragment_spread(
        self, node: FragmentSpreadNode, *_args: Any
    ) -> Dict[str, Any]:

        fragment_name = node.name.value

        log.debug(f"Start recursive fragment visit {fragment_name}")

        fragment_node = _get_fragment(self.document, fragment_name)

        fragment_result = parse_result_recursive(
            self.schema,
            self.document,
            fragment_node,
            self.current_result,
            visit_fragment=True,
        )

        log.debug(
            f"Result of recursive fragment visit {fragment_name}: {fragment_result}"
        )

        return cast(Dict[str, Any], fragment_result)

    @staticmethod
    def leave_inline_fragment(node: InlineFragmentNode, *_args: Any) -> Dict[str, Any]:

        selections = cast(List[Dict[str, Any]], node.selection_set)
        return {k: v for s in selections for k, v in s.items()}


def parse_result_recursive(
    schema: GraphQLSchema,
    document: DocumentNode,
    node: Node,
    result: Optional[Dict[str, Any]],
    initial_type: Optional[GraphQLType] = None,
    inside_list_level: int = 0,
    visit_fragment: bool = False,
    operation_name: Optional[str] = None,
) -> Any:

    if result is None:
        return None

    type_info = TypeInfo(schema, initial_type=initial_type)

    visited = visit(
        node,
        TypeInfoVisitor(
            type_info,
            ParseResultVisitor(
                schema,
                document,
                node,
                result,
                type_info=type_info,
                inside_list_level=inside_list_level,
                visit_fragment=visit_fragment,
                operation_name=operation_name,
            ),
        ),
        visitor_keys=RESULT_DOCUMENT_KEYS,
    )

    return visited


def parse_result(
    schema: GraphQLSchema,
    document: DocumentNode,
    result: Optional[Dict[str, Any]],
    operation_name: Optional[str] = None,
) -> Optional[Dict[str, Any]]:
    """Unserialize a result received from a GraphQL backend.

    :param schema: the GraphQL schema
    :param document: the document representing the query sent to the backend
    :param result: the serialized result received from the backend
    :param operation_name: the optional operation name

    :returns: a parsed result with scalars and enums parsed depending on
              their definition in the schema.

    Given a schema, a query and a serialized result,
    provide a new result with parsed values.

    If the result contains only built-in GraphQL scalars (String, Int, Float, ...)
    then the parsed result should be unchanged.

    If the result contains custom scalars or enums, then those values
    will be parsed with the parse_value method of the custom scalar or enum
    definition in the schema."""

    return parse_result_recursive(
        schema, document, document, result, operation_name=operation_name
    )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('./..'))


# -- Project information -----------------------------------------------------

project = 'gql 3'
copyright = '2020, graphql-python.org'
author = 'graphql-python.org'

# The full version, including alpha/beta/rc tags
from gql import __version__
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinxarg.ext',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx_rtd_theme'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Output file base name for HTML help builder.
htmlhelp_basename = 'gql-3-doc'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# -- AutoDoc configuration -------------------------------------------------
# autoclass_content = "both"
autodoc_default_options = {
    'members': True,
    'inherited-members': True,
    'special-members': '__init__',
    'undoc-members': True,
    'show-inheritance': True
}
autosummary_generate = True

# -- Intersphinx configuration ---------------------------------------------
intersphinx_mapping = {
    'aiohttp': ('https://docs.aiohttp.org/en/stable/', None),
    'graphql': ('https://graphql-core-3.readthedocs.io/en/latest/', None),
    'multidict': ('https://multidict.aio-libs.org/en/stable/', None),
    'python': ('https://docs.python.org/3/', None),
    'requests': ('https://requests.readthedocs.io/en/latest/', None),
    'websockets': ('https://websockets.readthedocs.io/en/11.0.3/', None),
    'yarl': ('https://yarl.aio-libs.org/en/stable/', None),
}

nitpick_ignore = [
    # graphql-core: should be fixed
    ('py:class', 'graphql.execution.execute.ExecutionResult'),
    ('py:class', 'Source'),
    ('py:class', 'GraphQLSchema'),

    # asyncio: should be fixed
    ('py:class', 'asyncio.locks.Event'),

    # aiohttp: should be fixed
    # See issue: https://github.com/aio-libs/aiohttp/issues/10468
    ('py:class', 'aiohttp.client.ClientSession'),
    ('py:class', 'aiohttp.client_reqrep.Fingerprint'),
    ('py:class', 'aiohttp.helpers.BasicAuth'),

    # multidict: should be fixed
    ('py:class', 'multidict._multidict.CIMultiDictProxy'),
    ('py:class', 'multidict._multidict.CIMultiDict'),
    ('py:class', 'multidict._multidict.istr'),

    # websockets: first bump websockets version
    ('py:class', 'websockets.datastructures.SupportsKeysAndGetItem'),
    ('py:class', 'websockets.typing.Subprotocol'),

    # httpx: no sphinx docs yet https://github.com/encode/httpx/discussions/3091
    ('py:class', 'httpx.AsyncClient'),
    ('py:class', 'httpx.Client'),
    ('py:class', 'httpx.Headers'),

    # botocore: no sphinx docs
    ('py:class', 'botocore.auth.BaseSigner'),
    ('py:class', 'botocore.awsrequest.AWSRequest'),
    ('py:class', 'botocore.credentials.Credentials'),
    ('py:class', 'botocore.session.Session'),

    # gql: ignore private classes
    ('py:class', 'gql.transport.httpx._HTTPXTransport'),
    ('py:class', 'gql.client._CallableT'),
]



import pytest

from gql import Client, gql
from gql.transport.exceptions import TransportConnectionFailed

from .conftest import get_localhost_ssl_context_client

# Marking all tests in this file with the websockets marker
pytestmark = pytest.mark.websockets

query1_str = """
    query getContinents {
      continents {
        code
        name
      }
    }
"""

default_query_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}},'
    '"status":"ok"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)


@pytest.fixture
def ws_server_helper(request):
    from .conftest import PhoenixChannelServerHelper

    yield PhoenixChannelServerHelper


async def query_server(ws):
    from .conftest import PhoenixChannelServerHelper

    await PhoenixChannelServerHelper.send_connection_ack(ws)
    await ws.recv()
    await ws.send(default_query_server_answer)
    await PhoenixChannelServerHelper.send_close(ws)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [query_server], indirect=True)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_phoenix_channel_query(server, query_str):
    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = PhoenixChannelWebsocketsTransport(channel_name="test_channel", url=url)

    query = gql(query_str)
    async with Client(transport=transport) as session:
        result = await session.execute(query)

    print("Client received:", result)
    continents = result["continents"]
    print("Continents received:", continents)
    africa = continents[0]
    assert africa["code"] == "AF"


@pytest.mark.asyncio
@pytest.mark.parametrize("ws_ssl_server", [query_server], indirect=True)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_phoenix_channel_query_ssl(ws_ssl_server, query_str):
    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    server = ws_ssl_server
    url = f"wss://{server.hostname}:{server.port}{path}"

    extra_args = {}

    _, ssl_context = get_localhost_ssl_context_client()

    extra_args["ssl"] = ssl_context

    transport = PhoenixChannelWebsocketsTransport(
        channel_name="test_channel",
        url=url,
        **extra_args,
    )

    query = gql(query_str)
    async with Client(transport=transport) as session:
        result = await session.execute(query)

    print("Client received:", result)


@pytest.mark.asyncio
@pytest.mark.parametrize("ws_ssl_server", [query_server], indirect=True)
@pytest.mark.parametrize("query_str", [query1_str])
@pytest.mark.parametrize("verify_https", ["explicitely_enabled", "default"])
async def test_phoenix_channel_query_ssl_self_cert_fail(
    ws_ssl_server, query_str, verify_https
):
    from ssl import SSLCertVerificationError

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    server = ws_ssl_server
    url = f"wss://{server.hostname}:{server.port}{path}"

    extra_args = {}

    if verify_https == "explicitely_enabled":
        extra_args["ssl"] = True

    transport = PhoenixChannelWebsocketsTransport(
        channel_name="test_channel",
        url=url,
        **extra_args,
    )

    query = gql(query_str)

    with pytest.raises(TransportConnectionFailed) as exc_info:
        async with Client(transport=transport) as session:
            await session.execute(query)

    cause = exc_info.value.__cause__

    assert isinstance(cause, SSLCertVerificationError)

    expected_error = "certificate verify failed: self-signed certificate"

    assert expected_error in str(cause)


query2_str = """
    subscription getContinents {
      continents {
        code
        name
      }
    }
"""

subscription_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":'
    '{"subscriptionId":"test_subscription"},'
    '"status":"ok"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

subscription_data_server_answer = (
    '{"event":"subscription:data","payload":'
    '{"subscriptionId":"test_subscription","result":'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}},'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

unsubscribe_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":'
    '{"subscriptionId":"test_subscription"},'
    '"status":"ok"},'
    '"ref":3,'
    '"topic":"test_topic"}'
)


async def subscription_server(ws):
    from .conftest import PhoenixChannelServerHelper

    await PhoenixChannelServerHelper.send_connection_ack(ws)
    await ws.recv()
    await ws.send(subscription_server_answer)
    await ws.send(subscription_data_server_answer)
    await ws.recv()
    await ws.send(unsubscribe_server_answer)
    # Unsubscribe will remove the listener
    # await PhoenixChannelServerHelper.send_close(ws)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [subscription_server], indirect=True)
@pytest.mark.parametrize("query_str", [query2_str])
async def test_phoenix_channel_subscription(server, query_str):
    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = PhoenixChannelWebsocketsTransport(channel_name="test_channel", url=url)

    first_result = None
    query = gql(query_str)
    async with Client(transport=transport) as session:
        generator = session.subscribe(query)
        async for result in generator:
            first_result = result
            break

        # Using aclose here to make it stop cleanly on pypy
        await generator.aclose()

    print("Client received:", first_result)



from typing import Any, Dict

import pytest
from graphql import GraphQLError

from gql import gql
from gql.utilities import parse_result
from tests.starwars.schema import StarWarsSchema


def test_hero_name_and_friends_query():
    query = gql(
        """
        query HeroNameAndFriendsQuery {
          hero {
            id
            friends {
              name
            }
            name
          }
        }
        """
    )

    result = {
        "hero": {
            "id": "2001",
            "friends": [
                {"name": "Luke Skywalker"},
                {"name": "Han Solo"},
                {"name": "Leia Organa"},
            ],
            "name": "R2-D2",
        }
    }

    parsed_result = parse_result(StarWarsSchema, query.document, result)

    assert result == parsed_result


def test_hero_name_and_friends_query_with_fragment():
    """Testing for issue #445"""

    query = gql(
        """
        query HeroNameAndFriendsQuery {
          hero {
            ...HeroSummary
            friends {
              name
            }
          }
        }
        fragment HeroSummary on Character {
          id
          name
        }
        """
    )

    result = {
        "hero": {
            "id": "2001",
            "friends": [
                {"name": "Luke Skywalker"},
                {"name": "Han Solo"},
                {"name": "Leia Organa"},
            ],
            "name": "R2-D2",
        }
    }

    parsed_result = parse_result(StarWarsSchema, query.document, result)

    assert result == parsed_result


def test_key_not_found_in_result():

    query = gql(
        """
        {
          hero {
            id
          }
        }
        """
    )

    # Backend returned an invalid result without the hero key
    # Should be impossible. In that case, we ignore the missing key
    result: Dict[str, Any] = {}

    parsed_result = parse_result(StarWarsSchema, query.document, result)

    assert result == parsed_result


def test_invalid_result_raise_error():

    query = gql(
        """
        {
          hero {
            id
          }
        }
        """
    )

    result = {"hero": 5}

    with pytest.raises(GraphQLError) as exc_info:

        parse_result(StarWarsSchema, query.document, result)

    assert "Invalid result for container of field id: 5" in str(exc_info)


def test_fragment():

    query = gql(
        """
        query UseFragment {
          luke: human(id: "1000") {
            ...HumanFragment
          }
          leia: human(id: "1003") {
            ...HumanFragment
          }
        }
        fragment HumanFragment on Human {
          name
          homePlanet
        }
        """
    )

    result = {
        "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
        "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
    }

    parsed_result = parse_result(StarWarsSchema, query.document, result)

    assert result == parsed_result


def test_fragment_not_found():

    query = gql(
        """
        query UseFragment {
          luke: human(id: "1000") {
            ...HumanFragment
          }
        }
        """
    )

    result = {
        "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
    }

    with pytest.raises(GraphQLError) as exc_info:

        parse_result(StarWarsSchema, query.document, result)

    assert 'Fragment "HumanFragment" not found in document!' in str(exc_info)


def test_return_none_if_result_is_none():

    query = gql(
        """
        query {
          hero {
           id
          }
        }
        """
    )

    result = None

    assert parse_result(StarWarsSchema, query.document, result) is None


def test_null_result_is_allowed():

    query = gql(
        """
        query {
          hero {
           id
          }
        }
        """
    )

    result = {"hero": None}

    parsed_result = parse_result(StarWarsSchema, query.document, result)

    assert result == parsed_result


def test_inline_fragment():

    query = gql(
        """
        query UseFragment {
          luke: human(id: "1000") {
            ... on Human {
              name
              homePlanet
            }
          }
        }
        """
    )

    result = {
        "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
    }

    parsed_result = parse_result(StarWarsSchema, query.document, result)

    assert result == parsed_result



import asyncio
from typing import cast

from graphql import (
    GraphQLArgument,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
    IntrospectionQuery,
    get_introspection_query,
    graphql_sync,
    print_schema,
)

from .fixtures import (
    create_review,
    get_characters,
    get_droid,
    get_friends,
    get_hero_async,
    get_human,
    reviews,
)

episode_enum = GraphQLEnumType(
    "Episode",
    {
        "NEWHOPE": GraphQLEnumValue(
            4,
            description="Released in 1977.",
        ),
        "EMPIRE": GraphQLEnumValue(
            5,
            description="Released in 1980.",
        ),
        "JEDI": GraphQLEnumValue(
            6,
            description="Released in 1983.",
        ),
    },
    description="One of the films in the Star Wars Trilogy",
)


human_type: GraphQLObjectType
droid_type: GraphQLObjectType

character_interface = GraphQLInterfaceType(
    "Character",
    lambda: {
        "id": GraphQLField(
            GraphQLNonNull(GraphQLString), description="The id of the character."
        ),
        "name": GraphQLField(GraphQLString, description="The name of the character."),
        "friends": GraphQLField(
            GraphQLList(character_interface),  # type: ignore
            description="The friends of the character,"
            " or an empty list if they have none.",
        ),
        "appearsIn": GraphQLField(
            GraphQLList(episode_enum), description="Which movies they appear in."
        ),
    },
    resolve_type=lambda character, _info, _type: {
        "Human": human_type.name,
        "Droid": droid_type.name,
    }[character.type],
    description="A character in the Star Wars Trilogy",
)

human_type = GraphQLObjectType(
    "Human",
    lambda: {
        "id": GraphQLField(
            GraphQLNonNull(GraphQLString),
            description="The id of the human.",
        ),
        "name": GraphQLField(
            GraphQLString,
            description="The name of the human.",
        ),
        "friends": GraphQLField(
            GraphQLList(character_interface),
            description="The friends of the human, or an empty list if they have none.",
            resolve=lambda human, _info: get_friends(human),
        ),
        "appearsIn": GraphQLField(
            GraphQLList(episode_enum),
            description="Which movies they appear in.",
        ),
        "homePlanet": GraphQLField(
            GraphQLString,
            description="The home planet of the human, or null if unknown.",
        ),
    },
    interfaces=[character_interface],
    description="A humanoid creature in the Star Wars universe.",
)

droid_type = GraphQLObjectType(
    "Droid",
    lambda: {
        "id": GraphQLField(
            GraphQLNonNull(GraphQLString),
            description="The id of the droid.",
        ),
        "name": GraphQLField(
            GraphQLString,
            description="The name of the droid.",
        ),
        "friends": GraphQLField(
            GraphQLList(character_interface),
            description="The friends of the droid, or an empty list if they have none.",
            resolve=lambda droid, _info: get_friends(droid),
        ),
        "appearsIn": GraphQLField(
            GraphQLList(episode_enum),
            description="Which movies they appear in.",
        ),
        "primaryFunction": GraphQLField(
            GraphQLString,
            description="The primary function of the droid.",
        ),
    },
    interfaces=[character_interface],
    description="A mechanical creature in the Star Wars universe.",
)

review_type = GraphQLObjectType(
    "Review",
    lambda: {
        "episode": GraphQLField(episode_enum, description="The movie"),
        "stars": GraphQLField(
            GraphQLNonNull(GraphQLInt),
            description="The number of stars this review gave, 1-5",
        ),
        "commentary": GraphQLField(
            GraphQLString, description="Comment about the movie"
        ),
    },
    description="Represents a review for a movie",
)

review_input_type = GraphQLInputObjectType(
    "ReviewInput",
    lambda: {
        "stars": GraphQLInputField(GraphQLInt, description="0-5 stars"),
        "commentary": GraphQLInputField(
            GraphQLString, description="Comment about the movie, optional"
        ),
        "deprecated_input_field": GraphQLInputField(
            GraphQLString,
            description="deprecated field example",
            deprecation_reason="deprecated for testing",
        ),
    },
    description="The input object sent when someone is creating a new review",
)

query_type = GraphQLObjectType(
    "Query",
    lambda: {
        "hero": GraphQLField(
            character_interface,
            args={
                "episode": GraphQLArgument(
                    episode_enum,
                    description="If omitted, returns the hero of the whole saga. If "
                    "provided, returns the hero of that particular episode.",
                )
            },
            resolve=lambda _source, _info, episode=None: get_hero_async(episode),
        ),
        "human": GraphQLField(
            human_type,
            args={
                "id": GraphQLArgument(
                    description="id of the human",
                    type_=GraphQLNonNull(GraphQLString),
                )
            },
            resolve=lambda _source, _info, id: get_human(id),
        ),
        "droid": GraphQLField(
            droid_type,
            args={
                "id": GraphQLArgument(
                    description="id of the droid",
                    type_=GraphQLNonNull(GraphQLString),
                )
            },
            resolve=lambda _source, _info, id: get_droid(id),
        ),
        "characters": GraphQLField(
            GraphQLList(character_interface),
            args={
                "ids": GraphQLArgument(
                    GraphQLList(GraphQLString),
                    description="list of character ids",
                )
            },
            resolve=lambda _source, _info, ids=None: get_characters(ids),
        ),
    },
)

mutation_type = GraphQLObjectType(
    "Mutation",
    lambda: {
        "createReview": GraphQLField(
            review_type,
            args={
                "episode": GraphQLArgument(
                    episode_enum,
                    description="Episode to create review",
                ),
                "review": GraphQLArgument(
                    description="set alive status",
                    type_=review_input_type,
                ),
            },
            resolve=lambda _source, _info, episode=None, review=None: create_review(
                episode, review
            ),
        ),
    },
    description="The mutation type, represents all updates we can make to our data",
)


async def subscribe_reviews(_root, _info, episode):
    for review in reviews[episode]:
        yield review
        await asyncio.sleep(0.1)


async def resolve_review(review, _info, **_args):
    return review


subscription_type = GraphQLObjectType(
    "Subscription",
    lambda: {
        "reviewAdded": GraphQLField(
            review_type,
            args={
                "episode": GraphQLArgument(
                    episode_enum,
                    description="Episode to review",
                )
            },
            subscribe=subscribe_reviews,
            resolve=resolve_review,
        )
    },
)


StarWarsSchema = GraphQLSchema(
    query=query_type,
    mutation=mutation_type,
    subscription=subscription_type,
    types=[human_type, droid_type, review_type, review_input_type],
)


StarWarsIntrospection = cast(
    IntrospectionQuery, graphql_sync(StarWarsSchema, get_introspection_query()).data
)

StarWarsTypeDef = print_schema(StarWarsSchema)



from .build_client_schema import build_client_schema
from .get_introspection_query_ast import get_introspection_query_ast
from .node_tree import node_tree
from .parse_result import parse_result
from .serialize_variable_values import serialize_value, serialize_variable_values
from .update_schema_enum import update_schema_enum
from .update_schema_scalars import update_schema_scalar, update_schema_scalars

__all__ = [
    "build_client_schema",
    "node_tree",
    "parse_result",
    "get_introspection_query_ast",
    "serialize_variable_values",
    "serialize_value",
    "update_schema_enum",
    "update_schema_scalars",
    "update_schema_scalar",
]



import asyncio

from gql import Client, gql
from gql.transport.phoenix_channel_websockets import PhoenixChannelWebsocketsTransport


async def main():

    transport = PhoenixChannelWebsocketsTransport(
        channel_name="YOUR_CHANNEL", url="wss://YOUR_URL/graphql"
    )

    # Using `async with` on the client will start a connection on the transport
    # and provide a `session` variable to execute queries on this connection
    async with Client(transport=transport) as session:

        # Execute single query
        query = gql(
            """
            query yourQuery {
                ...
            }
        """
        )

        result = await session.execute(query)
        print(result)


asyncio.run(main())



from datetime import datetime, timedelta
from typing import Any, Dict, Optional

from graphql.error import GraphQLError
from graphql.language import ValueNode
from graphql.pyutils import inspect
from graphql.type import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLList,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
)
from graphql.utilities import value_from_ast_untyped

from gql import Client, gql


def serialize_datetime(value: Any) -> str:
    if not isinstance(value, datetime):
        raise GraphQLError("Cannot serialize datetime value: " + inspect(value))
    return value.isoformat()


def parse_datetime_value(value: Any) -> datetime:

    if isinstance(value, str):
        try:
            # Note: a more solid custom scalar should use dateutil.parser.isoparse
            #       Not using it here in the test to avoid adding another dependency
            return datetime.fromisoformat(value)
        except Exception:
            raise GraphQLError("Cannot parse datetime value : " + inspect(value))

    else:
        raise GraphQLError("Cannot parse datetime value: " + inspect(value))


def parse_datetime_literal(
    value_node: ValueNode, variables: Optional[Dict[str, Any]] = None
) -> datetime:
    ast_value = value_from_ast_untyped(value_node, variables)
    if not isinstance(ast_value, str):
        raise GraphQLError("Cannot parse literal datetime value: " + inspect(ast_value))

    return parse_datetime_value(ast_value)


DatetimeScalar = GraphQLScalarType(
    name="Datetime",
    serialize=serialize_datetime,
    parse_value=parse_datetime_value,
    parse_literal=parse_datetime_literal,
)


def resolve_shift_days(root, _info, time, days):
    return time + timedelta(days=days)


def resolve_latest(root, _info, times):
    return max(times)


def resolve_seconds(root, _info, interval):
    print(f"interval={interval!r}")
    return (interval["end"] - interval["start"]).total_seconds()


IntervalInputType = GraphQLInputObjectType(
    "IntervalInput",
    fields={
        "start": GraphQLInputField(
            DatetimeScalar,
            default_value=datetime(2021, 11, 12, 11, 58, 13, 461161),
        ),
        "end": GraphQLInputField(DatetimeScalar),
    },
)

queryType = GraphQLObjectType(
    name="RootQueryType",
    fields={
        "shiftDays": GraphQLField(
            DatetimeScalar,
            args={
                "time": GraphQLArgument(DatetimeScalar),
                "days": GraphQLArgument(GraphQLInt),
            },
            resolve=resolve_shift_days,
        ),
        "latest": GraphQLField(
            DatetimeScalar,
            args={"times": GraphQLArgument(GraphQLList(DatetimeScalar))},
            resolve=resolve_latest,
        ),
        "seconds": GraphQLField(
            GraphQLInt,
            args={"interval": GraphQLArgument(IntervalInputType)},
            resolve=resolve_seconds,
        ),
    },
)

schema = GraphQLSchema(query=queryType)


def test_shift_days():

    client = Client(schema=schema, parse_results=True, serialize_variables=True)

    now = datetime.fromisoformat("2021-11-12T11:58:13.461161")

    query = gql("query shift5days($time: Datetime) {shiftDays(time: $time, days: 5)}")

    query.variable_values = {
        "time": now,
    }

    result = client.execute(query)

    print(result)

    assert result["shiftDays"] == datetime.fromisoformat("2021-11-17T11:58:13.461161")


def test_shift_days_serialized_manually_in_query():

    client = Client(schema=schema)

    query = gql(
        """{
        shiftDays(time: "2021-11-12T11:58:13.461161", days: 5)
    }"""
    )

    result = client.execute(query, parse_result=True)

    print(result)

    assert result["shiftDays"] == datetime.fromisoformat("2021-11-17T11:58:13.461161")


def test_shift_days_serialized_manually_in_variables():

    client = Client(schema=schema, parse_results=True)

    query = gql("query shift5days($time: Datetime) {shiftDays(time: $time, days: 5)}")

    query.variable_values = {
        "time": "2021-11-12T11:58:13.461161",
    }

    result = client.execute(query)

    print(result)

    assert result["shiftDays"] == datetime.fromisoformat("2021-11-17T11:58:13.461161")


def test_latest():

    client = Client(schema=schema, parse_results=True)

    now = datetime.fromisoformat("2021-11-12T11:58:13.461161")
    in_five_days = datetime.fromisoformat("2021-11-17T11:58:13.461161")

    query = gql("query latest($times: [Datetime!]!) {latest(times: $times)}")

    query.variable_values = {
        "times": [now, in_five_days],
    }

    result = client.execute(query, serialize_variables=True)

    print(result)

    assert result["latest"] == in_five_days


def test_seconds():
    client = Client(schema=schema)

    now = datetime.fromisoformat("2021-11-12T11:58:13.461161")
    in_five_days = datetime.fromisoformat("2021-11-17T11:58:13.461161")

    query = gql(
        "query seconds($interval: IntervalInput) {seconds(interval: $interval)}"
    )

    query.variable_values = {"interval": {"start": now, "end": in_five_days}}

    result = client.execute(query, serialize_variables=True)

    print(result)

    assert result["seconds"] == 432000


def test_seconds_omit_optional_start_argument():
    client = Client(schema=schema)

    in_five_days = datetime.fromisoformat("2021-11-17T11:58:13.461161")

    query = gql(
        "query seconds($interval: IntervalInput) {seconds(interval: $interval)}"
    )

    query.variable_values = {"interval": {"end": in_five_days}}

    result = client.execute(query, serialize_variables=True)

    print(result)

    assert result["seconds"] == 432000



from enum import Enum
from typing import Optional

import pytest
from graphql import (
    GraphQLArgument,
    GraphQLEnumType,
    GraphQLField,
    GraphQLList,
    GraphQLNamedType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
)

from gql import Client, gql
from gql.utilities import update_schema_enum


class Color(Enum):
    RED = 0
    GREEN = 1
    BLUE = 2
    YELLOW = 3
    CYAN = 4
    MAGENTA = 5


RED = Color.RED
GREEN = Color.GREEN
BLUE = Color.BLUE
YELLOW = Color.YELLOW
CYAN = Color.CYAN
MAGENTA = Color.MAGENTA

ALL_COLORS = [c for c in Color]

ColorType = GraphQLEnumType("Color", {c.name: c for c in Color})


def resolve_opposite(_root, _info, color):
    opposite_colors = {
        RED: CYAN,
        GREEN: MAGENTA,
        BLUE: YELLOW,
        YELLOW: BLUE,
        CYAN: RED,
        MAGENTA: GREEN,
    }

    return opposite_colors[color]


def resolve_all(_root, _info):
    return ALL_COLORS


list_of_list_of_list = [[[RED, GREEN], [GREEN, BLUE]], [[YELLOW, CYAN], [MAGENTA, RED]]]


def resolve_list_of_list_of_list(_root, _info):
    return list_of_list_of_list


def resolve_list_of_list(_root, _info):
    return list_of_list_of_list[0]


def resolve_list(_root, _info):
    return list_of_list_of_list[0][0]


queryType = GraphQLObjectType(
    name="RootQueryType",
    fields={
        "all": GraphQLField(
            GraphQLList(ColorType),
            resolve=resolve_all,
        ),
        "opposite": GraphQLField(
            ColorType,
            args={"color": GraphQLArgument(ColorType)},
            resolve=resolve_opposite,
        ),
        "list_of_list_of_list": GraphQLField(
            GraphQLNonNull(
                GraphQLList(
                    GraphQLNonNull(GraphQLList(GraphQLNonNull(GraphQLList(ColorType))))
                )
            ),
            resolve=resolve_list_of_list_of_list,
        ),
        "list_of_list": GraphQLField(
            GraphQLNonNull(GraphQLList(GraphQLNonNull(GraphQLList(ColorType)))),
            resolve=resolve_list_of_list,
        ),
        "list": GraphQLField(
            GraphQLNonNull(GraphQLList(ColorType)),
            resolve=resolve_list,
        ),
    },
)

schema = GraphQLSchema(query=queryType)


def test_parse_value_enum():

    result = ColorType.parse_value("RED")

    print(result)

    assert isinstance(result, Color)
    assert result is RED


def test_serialize_enum():

    result = ColorType.serialize(RED)

    print(result)

    assert result == "RED"


def test_get_all_colors():

    query = gql("{all}")

    client = Client(schema=schema, parse_results=True)

    result = client.execute(query)

    print(result)

    all_colors = result["all"]

    assert all_colors == ALL_COLORS


def test_opposite_color_literal():

    client = Client(schema=schema, parse_results=True)

    query = gql("{opposite(color: RED)}")

    result = client.execute(query)

    print(result)

    opposite_color = result["opposite"]

    assert isinstance(opposite_color, Color)
    assert opposite_color == CYAN


def test_opposite_color_variable_serialized_manually():

    client = Client(schema=schema, parse_results=True)

    query = gql(
        """
        query GetOppositeColor($color: Color) {
            opposite(color:$color)
        }"""
    )

    query.variable_values = {
        "color": "RED",
    }

    result = client.execute(query)

    print(result)

    opposite_color = result["opposite"]

    assert isinstance(opposite_color, Color)
    assert opposite_color == CYAN


def test_opposite_color_variable_serialized_by_gql():

    client = Client(schema=schema, parse_results=True)

    query = gql(
        """
        query GetOppositeColor($color: Color) {
            opposite(color:$color)
        }"""
    )

    query.variable_values = {
        "color": RED,
    }

    result = client.execute(query, serialize_variables=True)

    print(result)

    opposite_color = result["opposite"]

    assert isinstance(opposite_color, Color)
    assert opposite_color == CYAN


def test_list():

    query = gql("{list}")

    client = Client(schema=schema, parse_results=True)

    result = client.execute(query)

    print(result)

    big_list = result["list"]

    assert big_list == list_of_list_of_list[0][0]


def test_list_of_list():

    query = gql("{list_of_list}")

    client = Client(schema=schema, parse_results=True)

    result = client.execute(query)

    print(result)

    big_list = result["list_of_list"]

    assert big_list == list_of_list_of_list[0]


def test_list_of_list_of_list():

    query = gql("{list_of_list_of_list}")

    client = Client(schema=schema, parse_results=True)

    result = client.execute(query)

    print(result)

    big_list = result["list_of_list_of_list"]

    assert big_list == list_of_list_of_list


def test_update_schema_enum():

    color_type: Optional[GraphQLNamedType]

    color_type = schema.get_type("Color")
    assert isinstance(color_type, GraphQLEnumType)
    assert color_type is not None
    assert color_type.parse_value("RED") == Color.RED

    # Using values

    update_schema_enum(schema, "Color", Color, use_enum_values=True)

    color_type = schema.get_type("Color")
    assert isinstance(color_type, GraphQLEnumType)
    assert color_type is not None
    assert color_type.parse_value("RED") == 0
    assert color_type.serialize(1) == "GREEN"

    update_schema_enum(schema, "Color", Color)

    color_type = schema.get_type("Color")
    assert isinstance(color_type, GraphQLEnumType)
    assert color_type is not None
    assert color_type.parse_value("RED") == Color.RED
    assert color_type.serialize(Color.RED) == "RED"


def test_update_schema_enum_errors():

    with pytest.raises(KeyError) as exc_info:
        update_schema_enum(schema, "Corlo", Color)

    assert "Enum Corlo not found in schema!" in str(exc_info)

    with pytest.raises(TypeError) as exc_info2:
        update_schema_enum(schema, "Color", 6)  # type: ignore

    assert "Invalid type for enum values: " in str(exc_info2)

    with pytest.raises(TypeError) as exc_info3:
        update_schema_enum(schema, "RootQueryType", Color)

    assert 'The type "RootQueryType" is not a GraphQLEnumType, it is a' in str(
        exc_info3
    )

    with pytest.raises(KeyError) as exc_info4:
        update_schema_enum(schema, "Color", {"RED": Color.RED})

    assert 'Enum key "GREEN" not found in provided values!' in str(exc_info4)


def test_parse_results_with_operation_type():

    client = Client(schema=schema, parse_results=True)

    query = gql(
        """
        query GetAll {
            all
        }
        query GetOppositeColor($color: Color) {
            opposite(color:$color)
        }
        query GetOppositeColor2($color: Color) {
            other_opposite:opposite(color:$color)
        }
        query GetOppositeColor3 {
            opposite(color: YELLOW)
        }
        query GetListOfListOfList {
            list_of_list_of_list
        }
        """
    )

    query.variable_values = {
        "color": "RED",
    }
    query.operation_name = "GetOppositeColor"

    result = client.execute(query)

    print(result)

    opposite_color = result["opposite"]

    assert isinstance(opposite_color, Color)
    assert opposite_color == CYAN



import asyncio
import json
import logging
import signal as signal_module
import sys
import textwrap
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from typing import Any, Dict, Optional

from graphql import GraphQLError, print_schema
from yarl import URL

from gql import Client, __version__, gql
from gql.transport import AsyncTransport
from gql.transport.exceptions import TransportQueryError

description = """
Send GraphQL queries from the command line using http(s) or websockets.
If used interactively, write your query, then use Ctrl-D (EOF) to execute it.
"""

examples = """
EXAMPLES
========

# Simple query using https
echo 'query { continent(code:"AF") { name } }' | \
gql-cli https://countries.trevorblades.com

# Simple query using websockets
echo 'query { continent(code:"AF") { name } }' | \
gql-cli wss://countries.trevorblades.com/graphql

# Query with variable
echo 'query getContinent($code:ID!) { continent(code:$code) { name } }' | \
gql-cli https://countries.trevorblades.com --variables code:AF

# Interactive usage (insert your query in the terminal, then press Ctrl-D to execute it)
gql-cli wss://countries.trevorblades.com/graphql --variables code:AF

# Execute query saved in a file
cat query.gql | gql-cli wss://countries.trevorblades.com/graphql

# Print the schema of the backend
gql-cli https://countries.trevorblades.com/graphql --print-schema

"""


def positive_int_or_none(value_str: str) -> Optional[int]:
    """Convert a string argument value into either an int or None.

    Raise a ValueError if the argument is negative or a string which is not "none"
    """
    try:
        value_int = int(value_str)
    except ValueError:
        if value_str.lower() == "none":
            return None
        else:
            raise

    if value_int < 0:
        raise ValueError

    return value_int


def get_parser(with_examples: bool = False) -> ArgumentParser:
    """Provides an ArgumentParser for the gql-cli script.

    This function is also used by sphinx to generate the script documentation.

    :param with_examples: set to False by default so that the examples are not
                          present in the sphinx docs (they are put there with
                          a different layout)
    """

    parser = ArgumentParser(
        description=description,
        epilog=examples if with_examples else None,
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "server", help="the server url starting with http://, https://, ws:// or wss://"
    )
    parser.add_argument(
        "-V",
        "--variables",
        nargs="*",
        help="query variables in the form key:json_value",
    )
    parser.add_argument(
        "-H", "--headers", nargs="*", help="http headers in the form key:value"
    )
    parser.add_argument("--version", action="version", version=f"v{__version__}")
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-d",
        "--debug",
        help="print lots of debugging statements (loglevel==DEBUG)",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
    )
    group.add_argument(
        "-v",
        "--verbose",
        help="show low level messages (loglevel==INFO)",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
    )
    parser.add_argument(
        "-o",
        "--operation-name",
        help="set the operation_name value",
        dest="operation_name",
    )
    parser.add_argument(
        "--print-schema",
        help="get the schema from instrospection and print it",
        action="store_true",
        dest="print_schema",
    )
    parser.add_argument(
        "--schema-download",
        nargs="*",
        help=textwrap.dedent(
            """select the introspection query arguments to download the schema.
            Only useful if --print-schema is used.
            By default, it will:

             - request field descriptions
             - request deprecated input fields

            Possible options:

             - descriptions:false             for a compact schema without comments
             - input_value_deprecation:false  to omit deprecated input fields
             - specified_by_url:true
             - schema_description:true
             - directive_is_repeatable:true"""
        ),
        dest="schema_download",
    )
    parser.add_argument(
        "--execute-timeout",
        help="set the execute_timeout argument of the Client (default: 10)",
        type=positive_int_or_none,
        default=10,
        dest="execute_timeout",
    )
    parser.add_argument(
        "--transport",
        default="auto",
        choices=[
            "auto",
            "aiohttp",
            "httpx",
            "phoenix",
            "websockets",
            "aiohttp_websockets",
            "appsync_http",
            "appsync_websockets",
        ],
        help=(
            "select the transport. 'auto' by default: "
            "aiohttp or websockets depending on url scheme"
        ),
        dest="transport",
    )

    appsync_description = """
By default, for an AppSync backend, the IAM authentication is chosen.

If you want API key or JWT authentication, you can provide one of the
following arguments:"""

    appsync_group = parser.add_argument_group(
        "AWS AppSync options", description=appsync_description
    )

    appsync_auth_group = appsync_group.add_mutually_exclusive_group()

    appsync_auth_group.add_argument(
        "--api-key",
        help="Provide an API key for authentication",
        dest="api_key",
    )

    appsync_auth_group.add_argument(
        "--jwt",
        help="Provide an JSON Web token for authentication",
        dest="jwt",
    )

    return parser


def get_transport_args(args: Namespace) -> Dict[str, Any]:
    """Extract extra arguments necessary for the transport
    from the parsed command line args

    Will create a headers dict by splitting the colon
    in the --headers arguments

    :param args: parsed command line arguments
    """

    transport_args: Dict[str, Any] = {}

    # Parse the headers argument
    headers = {}
    if args.headers is not None:
        for header in args.headers:

            try:
                # Split only the first colon (throw a ValueError if no colon is present)
                header_key, header_value = header.split(":", 1)

                headers[header_key] = header_value

            except ValueError:
                raise ValueError(f"Invalid header: {header}")

    if args.headers is not None:
        transport_args["headers"] = headers

    return transport_args


def get_execute_args(args: Namespace) -> Dict[str, Any]:
    """Extract extra arguments necessary for the execute or subscribe
    methods from the parsed command line args

    Extract the operation_name

    Extract the variable_values from the --variables argument
    by splitting the first colon, then loads the json value,
    We try to add double quotes around the value if it does not work first
    in order to simplify the passing of simple string values
    (we allow --variables KEY:VALUE instead of KEY:\"VALUE\")

    :param args: parsed command line arguments
    """

    execute_args: Dict[str, Any] = {}

    # Parse the operation_name argument
    if args.operation_name is not None:
        execute_args["operation_name"] = args.operation_name

    # Parse the variables argument
    if args.variables is not None:

        variables = {}

        for var in args.variables:

            try:
                # Split only the first colon
                # (throw a ValueError if no colon is present)
                variable_key, variable_json_value = var.split(":", 1)

                # Extract the json value,
                # trying with double quotes if it does not work
                try:
                    variable_value = json.loads(variable_json_value)
                except json.JSONDecodeError:
                    try:
                        variable_value = json.loads(f'"{variable_json_value}"')
                    except json.JSONDecodeError:
                        raise ValueError

                # Save the value in the variables dict
                variables[variable_key] = variable_value

            except ValueError:
                raise ValueError(f"Invalid variable: {var}")

        execute_args["variable_values"] = variables

    return execute_args


def autodetect_transport(url: URL) -> str:
    """Detects which transport should be used depending on url."""

    if url.scheme in ["ws", "wss"]:
        try:
            import websockets  # noqa: F401

            transport_name = "websockets"
        except ImportError:  # pragma: no cover
            transport_name = "aiohttp_websockets"

    else:
        assert url.scheme in ["http", "https"]

        try:
            from gql.transport.aiohttp import AIOHTTPTransport  # noqa: F401

            transport_name = "aiohttp"
        except ModuleNotFoundError:  # pragma: no cover
            try:
                from gql.transport.httpx import HTTPXAsyncTransport  # noqa: F401

                transport_name = "httpx"
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "\n\nNo suitable dependencies has been found for an http(s) backend"
                    " (aiohttp or httpx).\n\n"
                    "Please check the install documentation at:\n"
                    "https://gql.readthedocs.io/en/stable/intro.html#installation\n"
                )

    return transport_name


def get_transport(args: Namespace) -> Optional[AsyncTransport]:
    """Instantiate a transport from the parsed command line arguments

    :param args: parsed command line arguments
    """

    # Get the url scheme from server parameter
    url = URL(args.server)

    # Validate scheme
    if url.scheme not in ["http", "https", "ws", "wss"]:
        raise ValueError("URL protocol should be one of: http, https, ws, wss")

    # Get extra transport parameters from command line arguments
    # (headers)
    transport_args = get_transport_args(args)

    # Either use the requested transport or autodetect it
    if args.transport == "auto":
        transport_name = autodetect_transport(url)
    else:
        transport_name = args.transport

    # Import the correct transport class depending on the transport name
    if transport_name == "aiohttp":
        from gql.transport.aiohttp import AIOHTTPTransport

        return AIOHTTPTransport(url=args.server, **transport_args)

    elif transport_name == "httpx":
        from gql.transport.httpx import HTTPXAsyncTransport

        return HTTPXAsyncTransport(url=args.server, **transport_args)

    elif transport_name == "phoenix":
        from gql.transport.phoenix_channel_websockets import (
            PhoenixChannelWebsocketsTransport,
        )

        return PhoenixChannelWebsocketsTransport(url=args.server, **transport_args)

    elif transport_name == "websockets":
        from gql.transport.websockets import WebsocketsTransport

        transport_args["ssl"] = url.scheme == "wss"

        return WebsocketsTransport(url=args.server, **transport_args)

    elif transport_name == "aiohttp_websockets":
        from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

        return AIOHTTPWebsocketsTransport(url=args.server, **transport_args)

    else:

        from gql.transport.appsync_auth import AppSyncAuthentication

        assert transport_name in ["appsync_http", "appsync_websockets"]
        assert url.host is not None

        auth: AppSyncAuthentication

        if args.api_key:
            from gql.transport.appsync_auth import AppSyncApiKeyAuthentication

            auth = AppSyncApiKeyAuthentication(host=url.host, api_key=args.api_key)

        elif args.jwt:
            from gql.transport.appsync_auth import AppSyncJWTAuthentication

            auth = AppSyncJWTAuthentication(host=url.host, jwt=args.jwt)

        else:
            from botocore.exceptions import NoRegionError

            from gql.transport.appsync_auth import AppSyncIAMAuthentication

            try:
                auth = AppSyncIAMAuthentication(host=url.host)
            except NoRegionError:
                # A warning message has been printed in the console
                return None

        transport_args["auth"] = auth

        if transport_name == "appsync_http":
            from gql.transport.aiohttp import AIOHTTPTransport

            return AIOHTTPTransport(url=args.server, **transport_args)

        else:
            from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

            try:
                return AppSyncWebsocketsTransport(url=args.server, **transport_args)
            except Exception:
                # This is for the NoCredentialsError but we cannot import it here
                return None


def get_introspection_args(args: Namespace) -> Dict:
    """Get the introspection args depending on the schema_download argument"""

    # Parse the headers argument
    introspection_args = {}

    possible_args = [
        "descriptions",
        "specified_by_url",
        "directive_is_repeatable",
        "schema_description",
        "input_value_deprecation",
    ]

    if args.schema_download is not None:
        for arg in args.schema_download:

            try:
                # Split only the first colon (throw a ValueError if no colon is present)
                arg_key, arg_value = arg.split(":", 1)

                if arg_key not in possible_args:
                    raise ValueError(f"Invalid schema_download: {args.schema_download}")

                arg_value = arg_value.lower()
                if arg_value not in ["true", "false"]:
                    raise ValueError(f"Invalid schema_download: {args.schema_download}")

                introspection_args[arg_key] = arg_value == "true"

            except ValueError:
                raise ValueError(f"Invalid schema_download: {args.schema_download}")

    return introspection_args


async def main(args: Namespace) -> int:
    """Main entrypoint of the gql-cli script

    :param args: The parsed command line arguments
    :return: The script exit code (0 = ok, 1 = error)
    """

    # Set requested log level
    if args.loglevel is not None:
        logging.basicConfig(level=args.loglevel)

    try:
        # Instantiate transport from command line arguments
        transport = get_transport(args)

        if transport is None:
            return 1

        # Get extra execute parameters from command line arguments
        # (variables, operation_name)
        execute_args = get_execute_args(args)

    except ValueError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except ModuleNotFoundError as e:  # pragma: no cover
        print(f"Error: {e}", file=sys.stderr)
        return 2

    # By default, the exit_code is 0 (everything is ok)
    exit_code = 0

    # Connect to the backend and provide a session
    async with Client(
        transport=transport,
        fetch_schema_from_transport=args.print_schema,
        introspection_args=get_introspection_args(args),
        execute_timeout=args.execute_timeout,
    ) as session:

        if args.print_schema:
            schema_str = print_schema(session.client.schema)
            print(schema_str)

            return exit_code

        while True:

            # Read multiple lines from input and trim whitespaces
            # Will read until EOF character is received (Ctrl-D)
            query_str = sys.stdin.read().strip()

            # Exit if query is empty
            if len(query_str) == 0:
                break

            # Parse query, continue on error
            try:
                query = gql(query_str)
            except GraphQLError as e:
                print(e, file=sys.stderr)
                exit_code = 1
                continue

            # Execute or Subscribe the query depending on transport
            try:
                try:
                    async for result in session.subscribe(query, **execute_args):
                        print(json.dumps(result))
                except KeyboardInterrupt:  # pragma: no cover
                    pass
                except NotImplementedError:
                    result = await session.execute(query, **execute_args)
                    print(json.dumps(result))
            except (GraphQLError, TransportQueryError) as e:
                print(e, file=sys.stderr)
                exit_code = 1

    return exit_code


def gql_cli() -> None:
    """Synchronously invoke ``main`` with the parsed command line arguments.

    Formerly ``scripts/gql-cli``, now registered as an ``entry_point``
    """
    # Get arguments from command line
    parser = get_parser(with_examples=True)
    args = parser.parse_args()

    try:
        # Create a new asyncio event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # Create a gql-cli task with the supplied arguments
        main_task = asyncio.ensure_future(main(args), loop=loop)

        # Add signal handlers to close gql-cli cleanly on Control-C
        for signal_name in ["SIGINT", "SIGTERM", "CTRL_C_EVENT", "CTRL_BREAK_EVENT"]:
            signal = getattr(signal_module, signal_name, None)

            if signal is None:
                continue

            try:
                loop.add_signal_handler(signal, main_task.cancel)
            except NotImplementedError:  # pragma: no cover
                # not all signals supported on all platforms
                pass

        # Run the asyncio loop to execute the task
        exit_code = 0
        try:
            exit_code = loop.run_until_complete(main_task)
        finally:
            loop.close()

        # Return with the correct exit code
        sys.exit(exit_code)
    except KeyboardInterrupt:  # pragma: no cover
        pass



import asyncio
import json
import logging
import os
import pathlib
import platform
import re
import ssl
import sys
import tempfile
import types
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Iterable, List, Union, cast

import pytest
import pytest_asyncio
from _pytest.fixtures import SubRequest

from gql import Client

all_transport_dependencies = ["aiohttp", "requests", "httpx", "websockets", "botocore"]


PyPy = platform.python_implementation() == "PyPy"


def pytest_addoption(parser):
    parser.addoption(
        "--run-online",
        action="store_true",
        default=False,
        help="run tests necessitating online resources",
    )
    for transport in all_transport_dependencies:
        parser.addoption(
            f"--{transport}-only",
            action="store_true",
            default=False,
            help=f"run tests necessitating only the {transport} dependency",
        )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "online: mark test as necessitating external online resources"
    )
    for transport in all_transport_dependencies:
        config.addinivalue_line(
            "markers",
            f"{transport}: mark test as necessitating the {transport} dependency",
        )


def pytest_collection_modifyitems(config, items):

    # --run-online given in cli: do not skip online tests
    if not config.getoption("--run-online"):
        skip_online = pytest.mark.skip(reason="need --run-online option to run")
        for item in items:
            if "online" in item.keywords:
                item.add_marker(skip_online)

    # --aiohttp-only
    # --requests-only
    # --httpx-only
    # --websockets-only
    for transport in all_transport_dependencies:

        other_transport_dependencies = [
            t for t in all_transport_dependencies if t != transport
        ]

        if config.getoption(f"--{transport}-only"):
            skip_transport = pytest.mark.skip(
                reason=f"need another dependency than {transport}"
            )
            for item in items:
                # Check if we have a dependency transport
                # other than the requested transport
                if any(t in item.keywords for t in other_transport_dependencies):
                    item.add_marker(skip_transport)


async def aiohttp_server_base(with_ssl=False):
    """Factory to create a TestServer instance, given an app.

    aiohttp_server(app, **kwargs)
    """
    from aiohttp.test_utils import TestServer as AIOHTTPTestServer

    servers = []

    async def go(app, *, port=None, **kwargs):  # type: ignore
        server = AIOHTTPTestServer(app, port=port)

        start_server_args = {**kwargs}
        if with_ssl:
            testcert, ssl_context = get_localhost_ssl_context()
            start_server_args["ssl"] = ssl_context

        await server.start_server(**start_server_args)
        servers.append(server)
        return server

    yield go

    while servers:
        await servers.pop().close()


@pytest_asyncio.fixture
async def aiohttp_server():
    async for server in aiohttp_server_base():
        yield server


@pytest_asyncio.fixture
async def ssl_aiohttp_server():
    async for server in aiohttp_server_base(with_ssl=True):
        yield server


# Adding debug logs
for name in [
    "websockets.legacy.server",
    "gql.transport.aiohttp",
    "gql.transport.aiohttp_websockets",
    "gql.transport.appsync",
    "gql.transport.common.base",
    "gql.transport.httpx",
    "gql.transport.phoenix_channel_websockets",
    "gql.transport.requests",
    "gql.transport.websockets",
    "gql.dsl",
    "gql.utilities.parse_result",
]:
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    if len(logger.handlers) < 1:
        logger.addHandler(logging.StreamHandler())

# Unit for timeouts. May be increased on slow machines by setting the
# GQL_TESTS_TIMEOUT_FACTOR environment variable.
# Copied from websockets source
MS = 0.001 * int(os.environ.get("GQL_TESTS_TIMEOUT_FACTOR", 1))


def get_localhost_ssl_context():
    # This is a copy of certificate from websockets tests folder
    #
    # Generate TLS certificate with:
    # $ openssl req -x509 -config test_localhost.cnf \
    #       -days 15340 -newkey rsa:2048 \
    #       -out test_localhost.crt -keyout test_localhost.key
    # $ cat test_localhost.key test_localhost.crt > test_localhost.pem
    # $ rm test_localhost.key test_localhost.crt
    testcert = bytes(pathlib.Path(__file__).with_name("test_localhost.pem"))
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    ssl_context.load_cert_chain(testcert)

    return (testcert, ssl_context)


def get_localhost_ssl_context_client():
    """
    Create a client-side SSL context that verifies the specific self-signed certificate
    used for our test.
    """
    # Get the certificate from the server setup
    cert_path = bytes(pathlib.Path(__file__).with_name("test_localhost_client.crt"))

    # Create client SSL context
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

    # Load just the certificate part as a trusted CA
    ssl_context.load_verify_locations(cafile=cert_path)

    # Require certificate verification
    ssl_context.verify_mode = ssl.CERT_REQUIRED

    # Enable hostname checking for localhost
    ssl_context.check_hostname = True

    return cert_path, ssl_context


class WebSocketServer:
    """Websocket server on localhost on a free port.

    This server allows us to test our client by simulating different correct and
    incorrect server responses.
    """

    def __init__(self, with_ssl: bool = False):
        self.with_ssl = with_ssl

    async def start(self, handler, extra_serve_args=None):

        import websockets

        print("Starting server")

        if extra_serve_args is None:
            extra_serve_args = {}

        if self.with_ssl:
            self.testcert, ssl_context = get_localhost_ssl_context()
            extra_serve_args["ssl"] = ssl_context

        # Adding dummy response headers
        extra_headers = {"dummy": "test1234"}

        def process_response(connection, request, response):
            response.headers.update(extra_headers)
            return response

        # Start a server with a random open port
        self.server = await websockets.serve(
            handler,
            "127.0.0.1",
            0,
            process_response=process_response,
            **extra_serve_args,
        )

        # Get hostname and port
        hostname, port = self.server.sockets[0].getsockname()[:2]  # type: ignore
        assert hostname == "127.0.0.1"

        self.hostname = hostname
        self.port = port

        print(f"Server started on port {port}")

    async def stop(self):
        print("Stopping server")

        self.server.close()
        try:
            await asyncio.wait_for(self.server.wait_closed(), timeout=5)
        except asyncio.TimeoutError:  # pragma: no cover
            pass

        print("Server stopped\n\n\n")


class AIOHTTPWebsocketServer:
    def __init__(self, with_ssl=False):
        self.runner = None
        self.site = None
        self.port = None
        self.hostname = "127.0.0.1"
        self.with_ssl = with_ssl
        self.ssl_context = None
        if with_ssl:
            _, self.ssl_context = get_localhost_ssl_context()

    def get_default_server_handler(answers: Iterable[str]) -> Callable:
        async def default_server_handler(request):

            import aiohttp
            import aiohttp.web
            from aiohttp import WSMsgType

            ws = aiohttp.web.WebSocketResponse()
            ws.headers.update({"dummy": "test1234"})
            await ws.prepare(request)

            try:
                # Init and ack
                msg = await ws.__anext__()
                assert msg.type == WSMsgType.TEXT
                result = msg.data
                json_result = json.loads(result)
                assert json_result["type"] == "connection_init"
                await ws.send_str('{"type":"connection_ack"}')
                query_id = 1

                # Wait for queries and send answers
                for answer in answers:
                    msg = await ws.__anext__()
                    if msg.type == WSMsgType.TEXT:
                        result = msg.data

                        print(f"Server received: {result}", file=sys.stderr)
                        if isinstance(answer, str) and "{query_id}" in answer:
                            answer_format_params = {"query_id": query_id}
                            formatted_answer = answer.format(**answer_format_params)
                        else:
                            formatted_answer = answer
                        await ws.send_str(formatted_answer)
                        await ws.send_str(
                            f'{{"type":"complete","id":"{query_id}","payload":null}}'
                        )
                        query_id += 1

                    elif msg.type == WSMsgType.ERROR:
                        print(f"WebSocket connection closed with: {ws.exception()}")
                        raise ws.exception()  # type: ignore
                    elif msg.type in (
                        WSMsgType.CLOSE,
                        WSMsgType.CLOSED,
                        WSMsgType.CLOSING,
                    ):
                        print("WebSocket connection closed")
                        raise ConnectionResetError

                # Wait for connection_terminate
                msg = await ws.__anext__()
                result = msg.data
                json_result = json.loads(result)
                assert json_result["type"] == "connection_terminate"

                # Wait for connection close
                msg = await ws.__anext__()

            except ConnectionResetError:
                pass

            except Exception as e:
                print(f"Server exception {e!s}", file=sys.stderr)

            await ws.close()
            return ws

        return default_server_handler

    async def shutdown_server(self, app):
        print("Shutting down server...")
        await app.shutdown()
        await app.cleanup()

    async def start(self, handler):
        import aiohttp
        import aiohttp.web

        app = aiohttp.web.Application()
        app.router.add_get("/graphql", handler)
        self.runner = aiohttp.web.AppRunner(app)
        await self.runner.setup()

        # Use port 0 to bind to an available port
        self.site = aiohttp.web.TCPSite(
            self.runner, self.hostname, 0, ssl_context=self.ssl_context
        )
        await self.site.start()

        # Retrieve the actual port the server is listening on
        assert self.site._server is not None
        sockets = self.site._server.sockets  # type: ignore
        if sockets:
            self.port = sockets[0].getsockname()[1]
            protocol = "https" if self.with_ssl else "http"
            print(f"Server started at {protocol}://{self.hostname}:{self.port}")

    async def stop(self):
        if self.site:
            await self.site.stop()
        if self.runner:
            await self.runner.cleanup()


@pytest_asyncio.fixture
async def aiohttp_ws_server(request):
    """Fixture used to start a dummy server to test the client behaviour
    using the aiohttp dependency.

    It can take as argument either a handler function for the websocket server for
    complete control OR an array of answers to be sent by the default server handler.
    """

    server_handler = get_aiohttp_ws_server_handler(request)

    try:
        test_server = AIOHTTPWebsocketServer()

        # Starting the server with the fixture param as the handler function
        await test_server.start(server_handler)

        yield test_server
    except Exception as e:
        print("Exception received in server fixture:", e)
    finally:
        await test_server.stop()


class WebSocketServerHelper:
    @staticmethod
    async def send_complete(ws, query_id):
        await ws.send(f'{{"type":"complete","id":"{query_id}","payload":null}}')

    @staticmethod
    async def send_keepalive(ws):
        await ws.send('{"type":"ka"}')

    @staticmethod
    async def send_ping(ws, payload=None):
        if payload is None:
            await ws.send('{"type":"ping"}')
        else:
            await ws.send(json.dumps({"type": "ping", "payload": payload}))

    @staticmethod
    async def send_pong(ws, payload=None):
        if payload is None:
            await ws.send('{"type":"pong"}')
        else:
            await ws.send(json.dumps({"type": "pong", "payload": payload}))

    @staticmethod
    async def send_connection_ack(ws, payload=None):

        # Line return for easy debugging
        print("")

        # Wait for init
        result = await ws.recv()
        json_result = json.loads(result)
        assert json_result["type"] == "connection_init"

        # Send ack
        if payload is None:
            await ws.send('{"type":"connection_ack"}')
        else:
            await ws.send(json.dumps({"type": "connection_ack", "payload": payload}))

    @staticmethod
    async def wait_connection_terminate(ws):
        result = await ws.recv()
        json_result = json.loads(result)
        assert json_result["type"] == "connection_terminate"


class PhoenixChannelServerHelper:
    @staticmethod
    async def send_close(ws):
        await ws.send('{"event":"phx_close"}')

    @staticmethod
    async def send_connection_ack(ws):

        # Line return for easy debugging
        print("")

        # Wait for init
        result = await ws.recv()
        json_result = json.loads(result)
        assert json_result["event"] == "phx_join"

        # Send ack
        await ws.send('{"event":"phx_reply", "payload": {"status": "ok"}, "ref": 1}')


class TemporaryFile:
    """Class used to generate temporary files for the tests"""

    def __init__(self, content: Union[str, bytearray, bytes]):

        mode = "w" if isinstance(content, str) else "wb"

        # We need to set the newline to '' so that the line returns
        # are not replaced by '\r\n' on windows
        newline = "" if isinstance(content, str) else None

        self.file = tempfile.NamedTemporaryFile(
            mode=mode, newline=newline, delete=False
        )

        with self.file as f:
            f.write(content)

    @property
    def filename(self):
        return self.file.name

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        os.unlink(self.filename)


def get_aiohttp_ws_server_handler(
    request: SubRequest,
) -> Callable:
    """Get the server handler for the aiohttp websocket server.

    Either get it from test or use the default server handler
    if the test provides only an array of answers.
    """

    server_handler: Callable

    if isinstance(request.param, types.FunctionType):
        server_handler = request.param

    else:
        answers = cast(List[str], request.param)
        server_handler = AIOHTTPWebsocketServer.get_default_server_handler(answers)

    return server_handler


def get_server_handler(
    request: SubRequest,
) -> Callable:
    """Get the server handler.

    Either get it from test or use the default server handler
    if the test provides only an array of answers.
    """

    from websockets.exceptions import ConnectionClosed

    if isinstance(request.param, types.FunctionType):
        server_handler: Callable = request.param

    else:
        answers = request.param

        async def default_server_handler(ws):

            try:
                await WebSocketServerHelper.send_connection_ack(ws)
                query_id = 1

                for answer in answers:
                    result = await ws.recv()
                    print(f"Server received: {result}", file=sys.stderr)

                    if isinstance(answer, str) and "{query_id}" in answer:
                        answer_format_params = {"query_id": query_id}
                        formatted_answer = answer.format(**answer_format_params)
                    else:
                        formatted_answer = answer

                    await ws.send(formatted_answer)
                    await WebSocketServerHelper.send_complete(ws, query_id)
                    query_id += 1

                await WebSocketServerHelper.wait_connection_terminate(ws)
                await ws.wait_closed()
            except ConnectionClosed:
                pass

        server_handler = default_server_handler

    return server_handler


@pytest_asyncio.fixture
async def ws_ssl_server(request):
    """Websockets server fixture using SSL.

    It can take as argument either a handler function for the websocket server for
    complete control OR an array of answers to be sent by the default server handler.
    """

    server_handler = get_server_handler(request)

    try:
        test_server = WebSocketServer(with_ssl=True)

        # Starting the server with the fixture param as the handler function
        await test_server.start(server_handler)

        yield test_server
    except Exception as e:
        print("Exception received in ws server fixture:", e)
    finally:
        await test_server.stop()


@pytest_asyncio.fixture
async def server(request):
    """Fixture used to start a dummy server to test the client behaviour.

    It can take as argument either a handler function for the websocket server for
    complete control OR an array of answers to be sent by the default server handler.
    """

    server_handler = get_server_handler(request)

    try:
        test_server = WebSocketServer()

        # Starting the server with the fixture param as the handler function
        await test_server.start(server_handler)

        yield test_server
    except Exception as e:
        print("Exception received in server fixture:", e)
    finally:
        await test_server.stop()


@pytest_asyncio.fixture
async def graphqlws_server(request):
    """Fixture used to start a dummy server with the graphql-ws protocol.

    Similar to the server fixture above but will return "graphql-transport-ws"
    as the server subprotocol.

    It can take as argument either a handler function for the websocket server for
    complete control OR an array of answers to be sent by the default server handler.
    """

    subprotocol = "graphql-transport-ws"

    server_handler = get_server_handler(request)

    try:
        test_server = WebSocketServer()

        # Starting the server with the fixture param as the handler function
        await test_server.start(
            server_handler, extra_serve_args={"subprotocols": [subprotocol]}
        )

        yield test_server
    except Exception as e:
        print("Exception received in server fixture:", e)
    finally:
        await test_server.stop()


@pytest_asyncio.fixture
async def client_and_server(server):
    """Helper fixture to start a server and a client connected to its port."""

    from gql.transport.websockets import WebsocketsTransport

    # Generate transport to connect to the server fixture
    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = WebsocketsTransport(url=url)

    async with Client(transport=transport) as session:

        # Yield both client session and server
        yield session, server


@pytest_asyncio.fixture
async def aiohttp_client_and_server(server):
    """
    Helper fixture to start a server and a client connected to its port
    with an aiohttp websockets transport.
    """

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    # Generate transport to connect to the server fixture
    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = AIOHTTPWebsocketsTransport(url=url)

    async with Client(transport=transport) as session:

        # Yield both client session and server
        yield session, server


@pytest_asyncio.fixture
async def aiohttp_client_and_aiohttp_ws_server(aiohttp_ws_server):
    """
    Helper fixture to start an aiohttp websocket server and
    a client connected to its port with an aiohttp websockets transport.
    """

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    server = aiohttp_ws_server

    # Generate transport to connect to the server fixture
    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = AIOHTTPWebsocketsTransport(url=url)

    async with Client(transport=transport) as session:

        # Yield both client session and server
        yield session, server


@pytest_asyncio.fixture
async def client_and_graphqlws_server(graphqlws_server):
    """Helper fixture to start a server with the graphql-ws prototocol
    and a client connected to its port."""

    from gql.transport.websockets import WebsocketsTransport

    # Generate transport to connect to the server fixture
    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = WebsocketsTransport(
        url=url,
        subprotocols=[WebsocketsTransport.GRAPHQLWS_SUBPROTOCOL],
    )

    async with Client(transport=transport) as session:

        # Yield both client session and server
        yield session, graphqlws_server


@pytest_asyncio.fixture
async def client_and_aiohttp_websocket_graphql_server(graphqlws_server):
    """Helper fixture to start a server with the graphql-ws prototocol
    and a client connected to its port."""

    from gql.transport.aiohttp_websockets import AIOHTTPWebsocketsTransport

    # Generate transport to connect to the server fixture
    path = "/graphql"
    url = f"ws://{graphqlws_server.hostname}:{graphqlws_server.port}{path}"
    transport = AIOHTTPWebsocketsTransport(
        url=url,
        subprotocols=[AIOHTTPWebsocketsTransport.GRAPHQLWS_SUBPROTOCOL],
    )

    async with Client(transport=transport) as session:

        # Yield both client session and server
        yield session, graphqlws_server


@pytest_asyncio.fixture
async def run_sync_test():
    async def run_sync_test_inner(server, test_function):
        """This function will run the test in a different Thread.

        This allows us to run sync code while aiohttp server can still run.
        """
        event_loop = asyncio.get_running_loop()
        executor = ThreadPoolExecutor(max_workers=2)
        test_task = event_loop.run_in_executor(executor, test_function)

        await test_task

        if hasattr(server, "close"):
            await server.close()

    return run_sync_test_inner


pytest_plugins = [
    "tests.fixtures.aws.fake_credentials",
    "tests.fixtures.aws.fake_request",
    "tests.fixtures.aws.fake_session",
    "tests.fixtures.aws.fake_signer",
]


def strip_braces_spaces(s):
    """Allow to ignore differences in graphql-core syntax between versions"""

    # Strip spaces after starting braces
    strip_front = s.replace("{ ", "{")

    # Strip spaces before closing braces only if one space is present
    strip_back = re.sub(r"([^\s]) }", r"\1}", strip_front)

    return strip_back


def make_upload_handler(
    nb_files=1,
    filenames=None,
    request_headers=None,
    file_headers=None,
    binary=False,
    expected_contents=None,
    expected_operations=None,
    expected_map=None,
    server_answer='{"data":{"success":true}}',
):
    assert expected_contents is not None
    assert expected_operations is not None
    assert expected_map is not None

    async def single_upload_handler(request):
        from aiohttp import web

        reader = await request.multipart()

        if request_headers is not None:
            for k, v in request_headers.items():
                assert request.headers[k] == v

        field_0 = await reader.next()
        assert field_0.name == "operations"
        field_0_text = await field_0.text()
        assert strip_braces_spaces(field_0_text) == expected_operations

        field_1 = await reader.next()
        assert field_1.name == "map"
        field_1_text = await field_1.text()
        assert field_1_text == expected_map

        for i in range(nb_files):
            field = await reader.next()
            assert field.name == str(i)
            if filenames is not None:
                assert field.filename == filenames[i]

            if binary:
                field_content = await field.read()
                assert field_content == expected_contents[i]
            else:
                field_text = await field.text()
                assert field_text == expected_contents[i]

            if file_headers is not None:
                for k, v in file_headers[i].items():
                    assert field.headers[k] == v

        final_field = await reader.next()
        assert final_field is None

        return web.Response(text=server_answer, content_type="application/json")

    return single_upload_handler



import warnings
from typing import Any, Dict, Optional, Union

from graphql import DocumentNode, GraphQLSchema, Source, parse, print_ast


class GraphQLRequest:
    """GraphQL Request to be executed."""

    def __init__(
        self,
        request: Union[DocumentNode, "GraphQLRequest", str],
        *,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
    ):
        """Initialize a GraphQL request.

        :param request: GraphQL request as DocumentNode object or as a string.
             If string, it will be converted to DocumentNode.
        :param variable_values: Dictionary of input parameters (Default: None).
        :param operation_name: Name of the operation that shall be executed.
            Only required in multi-operation documents (Default: None).
        :return: a :class:`GraphQLRequest <gql.GraphQLRequest>`
                 which can be later executed or subscribed by a
                 :class:`Client <gql.client.Client>`, by an
                 :class:`async session <gql.client.AsyncClientSession>` or by a
                 :class:`sync session <gql.client.SyncClientSession>`
        :raises graphql.error.GraphQLError: if a syntax error is encountered.
        """
        if isinstance(request, str):
            source = Source(request, "GraphQL request")
            self.document = parse(source)
        elif isinstance(request, DocumentNode):
            self.document = request
        elif not isinstance(request, GraphQLRequest):
            raise TypeError(f"Unexpected type for GraphQLRequest: {type(request)}")

        if isinstance(request, GraphQLRequest):
            self.document = request.document
            if variable_values is None:
                variable_values = request.variable_values
            if operation_name is None:
                operation_name = request.operation_name

        self.variable_values: Optional[Dict[str, Any]] = variable_values
        self.operation_name: Optional[str] = operation_name

    def serialize_variable_values(self, schema: GraphQLSchema) -> "GraphQLRequest":

        from .utilities.serialize_variable_values import serialize_variable_values

        assert self.variable_values

        return GraphQLRequest(
            self.document,
            variable_values=serialize_variable_values(
                schema=schema,
                document=self.document,
                variable_values=self.variable_values,
                operation_name=self.operation_name,
            ),
            operation_name=self.operation_name,
        )

    @property
    def payload(self) -> Dict[str, Any]:
        query_str = print_ast(self.document)
        payload: Dict[str, Any] = {"query": query_str}

        if self.operation_name:
            payload["operationName"] = self.operation_name

        if self.variable_values:
            payload["variables"] = self.variable_values

        return payload

    def __str__(self):
        return str(self.payload)


def support_deprecated_request(
    request: Union[GraphQLRequest, DocumentNode],
    kwargs: Dict,
) -> GraphQLRequest:
    """This methods is there temporarily to convert the old style of calling
    execute and subscribe methods with a DocumentNode,
    variable_values and operation_name arguments.
    """

    if isinstance(request, DocumentNode):
        warnings.warn(
            (
                "Using a DocumentNode is deprecated. Please use a "
                "GraphQLRequest instead."
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        request = GraphQLRequest(request)

    if not isinstance(request, GraphQLRequest):
        raise TypeError("request should be a GraphQLRequest object")

    variable_values = kwargs.pop("variable_values", None)
    operation_name = kwargs.pop("operation_name", None)

    if variable_values or operation_name:
        warnings.warn(
            (
                "Using variable_values and operation_name arguments of "
                "execute and subscribe methods is deprecated. Instead, "
                "please use the variable_values and operation_name properties "
                "of GraphQLRequest"
            ),
            DeprecationWarning,
            stacklevel=2,
        )

        request.variable_values = variable_values
        request.operation_name = operation_name

    return request
